# -*- coding: UTF-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import meta
import string  # pylint: disable=W0402
import re  
import test
import shutil, os


class TranslationIntegrityTestsMixin(object):
    ''' Unittests for translations. This class is subclassed below for each
        translated string in each language. '''
    
    conversionSpecificationRE = re.compile('%\(\w+\)[sd]')
    
    @staticmethod
    def findMatches(regex, search_string):
        matches = set()
        for match in re.findall(regex, search_string):
            matches.add(match)
        return matches
            
    def testMatchingConversionSpecifications(self):
        regex = self.conversionSpecificationRE
        matches_english = self.findMatches(regex, self.englishString)
        matches_translation = self.findMatches(regex, self.translatedString)
        self.assertEqual(matches_english, matches_translation, self.englishString)

    def testMatchingNonLiterals(self):
        for symbol in '\t', '|', '%s', '%d', '%.2f':
            self.assertEqual(self.englishString.count(symbol), 
                self.translatedString.count(symbol),
                "Symbol ('%s') doesn't match for '%s' and '%s'" % (symbol,
                    self.englishString, self.translatedString))
            
    def testMatchingAmpersands(self):
        # If the original string contains zero or one ampersands, it may be 
        # an accelerator. In that case, we don't require the translated string
        # to have an accelerator as well, because many translators don't use 
        # it and it doesn't break the application. However, if the original
        # string contains more than one ampersand it's probably HTML. In that
        # case we do require the number of ampersands to match exactly in the 
        # original and translated string.
        translatedString = self.removeUmlauts(self.translatedString)
        nrEnglishAmpersand = self.englishString.count('&')
        nrTranslatedAmpersand = translatedString.count('&')
        if nrEnglishAmpersand <= 1 and not '\n' in self.englishString:
            self.failUnless(nrTranslatedAmpersand in [0, 1], 
                "'%s' has more than one '&'" % self.translatedString)
        else:
            self.assertEqual(nrEnglishAmpersand, nrTranslatedAmpersand,
                "'%s' has more or less '&'s than '%s'" % (self.translatedString,
                self.englishString))

    usedShortcuts = dict()
    # Some keyboard shortcuts are used more than once, list those here:
    maxShortcuts = {'Ctrl-RETURN': 2, 'Shift+Ctrl+T': 3}
    
    def testUniqueShortCut(self):
        if '\t' in self.translatedString:
            shortcut = self.translatedString.split('\t')[1]
            shortcutKey = shortcut, self.language
            timesUsed = self.usedShortcuts.get(shortcutKey, 0)
            timesAllowed = self.maxShortcuts.get(shortcut, 1)
            self.failIf(timesUsed > timesAllowed, "Shortcut ('%s') used more "
                        "than once in language %s." % shortcutKey)
            self.usedShortcuts[shortcutKey] = timesUsed + 1
            
    def testMatchingShortCut(self):
        for shortcutPrefix in ('Ctrl+', 'Ctrl-', 'Shift+', 'Shift-', 
                               'Alt+', 'Alt-', 'Shift+Ctrl+', 'Shift-Ctrl-',
                               'Shift+Alt+', 'Shift-Alt-'):
            self.assertEqual(self.englishString.count('\t' + shortcutPrefix),
                             self.translatedString.count('\t' + shortcutPrefix),
                             "Shortcut prefix ('%s') doesn't match for '%s' "
                             "and '%s'" % (shortcutPrefix, self.englishString, 
                                           self.translatedString))
            
    def testShortCutIsAscii(self):
        ''' Test that the translated short cut key is using ASCII only. '''
        if '\t' in self.translatedString:
            shortcut = set(self.translatedString.split('\t')[1])
            self.failUnless(shortcut & set(string.ascii_letters + string.digits))
            
    @staticmethod
    def ellipsisCount(text):
        return text.count('...') + text.count('…')
    
    def testMatchingEllipses(self):
        self.assertEqual(self.ellipsisCount(self.englishString),
                         self.ellipsisCount(self.translatedString),
                         "Ellipses ('...') don't match for '%s' and '%s'" % \
                         (self.englishString, self.translatedString))

    umlautRE = re.compile(r'&[A-Za-z]uml;')
    
    @classmethod
    def removeUmlauts(cls, text):
        return re.sub(cls.umlautRE, '', text)      


class TranslationCoverageTestsMixin(object):
    def testNotComplete(self):
        if self.enabled:
            percentDone = 100.0 * len(self.translation) / len(self.strings)
            self.assertGreaterEqual(percentDone, 90.0, 'Translation for %s is only %.2f%% complete' % (self.language, percentDone))

    def testComplete(self):
        if not self.enabled:
            percentDone = 100.0 * len(self.translation) / len(self.strings)
            self.assertLess(percentDone, 90.0, 'Translation for %s is %.2f%% complete but disabled' % (self.language, percentDone))


def installAllTestCaseClasses():
    shutil.copyfile(os.path.join(os.path.dirname(__file__), '../../i18n.in/messages.pot'), 'messages.po')
    from taskcoachlib.i18n import po2dict
    po2dict.make('messages')
    allStrings = set(po2dict.STRINGS)
    for language, enabled in getLanguages():
        installTestCaseClasses(language, enabled, allStrings)


def getLanguages():
    return [(language, enabled) for language, enabled in meta.data.languages.values() \
            if language is not None]


def installTestCaseClasses(language, enabled, allStrings):
    translation = __import__('taskcoachlib.i18n.%s' % language, 
                             fromlist=['dict'])
    for englishString, translatedString in translation.dict.iteritems():        
        installTranslationTestCaseClass(language, englishString, 
                                              translatedString)
    installLanguageTestCaseClass(language, enabled, translation, allStrings)


def installTranslationTestCaseClass(language, englishString, 
                                          translatedString):
    testCaseClassName = translationTestCaseClassName(language, englishString)
    testCaseClass = translationTestCaseClass(testCaseClassName, 
        language, englishString, translatedString)
    globals()[testCaseClassName] = testCaseClass

def installLanguageTestCaseClass(language, enabled, translation, allStrings):
    testCaseClassName = languageTestCaseClassName(language)
    testCaseClass = languageTestCaseClass(testCaseClassName, language, enabled, translation, allStrings)
    globals()[testCaseClassName] = testCaseClass

def translationTestCaseClassName(language, englishString, 
                                 prefix='TranslationIntegrityTest'):
    ''' Generate a class name for the test case class based on the language
        and the English string. '''
    # Make sure we only use characters allowed in Python identifiers:
    englishString = englishString.replace(' ', '_')
    allowableCharacters = string.ascii_letters + string.digits + '_'
    englishString = ''.join([char for char in englishString \
                             if char in allowableCharacters])
    className = '%s_%s_%s' % (prefix, language, englishString)
    count = 0
    while className in globals():  # Make sure className is unique
        count += 1
        className = '%s_%s_%s_%d' % (prefix, language, englishString, count)
    return className

def languageTestCaseClassName(language, prefix='TranslationCoverageTests'):
    className = '%s_%s' % (prefix, language)
    count = 0
    while className in globals():
        count += 1
        className = '%s_%s_%s' % (prefix, language, count)
    return className

def translationTestCaseClass(className, language, englishString, translatedString):
    class_ = type(className, (TranslationIntegrityTestsMixin, test.TestCase), 
                  {})
    class_.language = language
    class_.englishString = englishString
    class_.translatedString = translatedString
    return class_

def languageTestCaseClass(className, language, enabled, translation, allStrings):
    class_ = type(className, (TranslationCoverageTestsMixin, test.TestCase), dict())
    class_.translation = translation.dict
    class_.enabled = enabled
    class_.strings = allStrings
    class_.language = language
    return class_


# Create all test cases and install them in the global name space:
installAllTestCaseClasses() 



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import persistence, config
from taskcoachlib.domain import task, category
from unittests import dummy
import test
from taskcoachlib.changes import ChangeMonitor


class DummyFile(object):
    name = 'testfile.tsk'
    encoding = 'utf-8'

    def close(self, *args, **kwargs):
        pass

    def write(self, *args, **kwargs):
        pass
    
    
class DummyTaskFile(persistence.TaskFile):
    def __init__(self, *args, **kwargs):
        self.saveCalled = 0
        self._throw = False
        super(DummyTaskFile, self).__init__(*args, **kwargs)

    def _read(self, *args, **kwargs):  # pylint: disable=W0613,W0221
        if self._throw:
            raise IOError
        else:
            return [task.Task()], [category.Category('category')], [], None, {self.monitor().guid(): self.monitor()}, None

    def exists(self, *args, **kwargs):  # pylint: disable=W0613
        return True
        
    def _openForRead(self, *args, **kwargs):  # pylint: disable=W0613
        return DummyFile()
        
    def _openForWrite(self, *args, **kwargs):  # pylint: disable=W0613
        return DummyFile()
    
    def save(self, *args, **kwargs):
        if kwargs.get('doNotify', True):
            self.saveCalled += 1
        super(DummyTaskFile, self).save(*args, **kwargs)

    def load(self, filename=None, throw=False, *args, **kwargs):  # pylint: disable=W0221
        self._throw = throw  # pylint: disable=W0201
        return super(DummyTaskFile, self).load(filename, *args, **kwargs)


class AutoSaverTestCase(test.TestCase):
    def setUp(self):
        task.Task.settings = self.settings = config.Settings(load=False)
        self.taskFile = DummyTaskFile()
        self.autoSaver = persistence.AutoSaver(self.settings)
        
    def tearDown(self):
        super(AutoSaverTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()
        del self.autoSaver # Make sure AutoSaver is not observing task files
        
    def testCreate(self):
        self.failIf(self.taskFile.saveCalled)
        
    def testFileChanged_ButNoFilenameAndAutoSaveOff(self):
        self.taskFile.tasks().append(task.Task())
        self.autoSaver.on_idle(dummy.Event())
        self.failIf(self.taskFile.saveCalled)
        
    def testFileChanged_ButAutoSaveOff(self):
        self.settings.set('file', 'autosave', 'False')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.tasks().append(task.Task())
        self.autoSaver.on_idle(dummy.Event())
        self.failIf(self.taskFile.saveCalled)
                
    def testFileChanged_ButNoFilename(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.tasks().append(task.Task())
        self.autoSaver.on_idle(dummy.Event())
        self.failIf(self.taskFile.saveCalled)
        
    def testFileChanged(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.tasks().append(task.Task())
        self.autoSaver.on_idle(dummy.Event())
        self.assertEqual(1, self.taskFile.saveCalled)
        
    def testSaveAsDoesNotTriggerAutoSave(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.saveas('newfilename.tsk')
        self.autoSaver.on_idle(dummy.Event())
        self.assertEqual(1, self.taskFile.saveCalled)
              
    def testCloseDoesNotTriggerAutoSave(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.tasks().append(task.Task())
        self.autoSaver.on_idle(dummy.Event())
        self.taskFile.close()
        self.assertEqual(1, self.taskFile.saveCalled)
        
    def testLoadDoesNotTriggerAutoSave(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.load()
        self.autoSaver.on_idle(dummy.Event())
        self.failIf(self.taskFile.saveCalled)

    def testLoadWithExceptionDoesNotTriggerAutoSave(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        try:
            self.taskFile.load(throw=True)
        except IOError:
            pass
        self.autoSaver.on_idle(dummy.Event())
        self.failIf(self.taskFile.saveCalled)
        
    def testMergeDoesTriggerAutoSave(self):
        self.settings.set('file', 'autosave', 'True')
        self.taskFile.setFilename('whatever.tsk')
        self.taskFile.merge('another-non-existing-file.tsk')
        self.autoSaver.on_idle(dummy.Event())
        self.assertEqual(1, self.taskFile.saveCalled)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.syncml.core import *
from taskcoachlib.i18n import _


class BaseSource(SyncSource):
    def __init__(self, callback, objectList, *args, **kwargs):
        super(BaseSource, self).__init__(*args, **kwargs)

        self.callback = callback
        self.objectList = objectList

        self.allObjectsList = [obj for obj in objectList]
        self.newObjectsList = [obj for obj in objectList if obj.isNew()]
        self.changedObjectsList = [obj for obj in objectList if obj.isModified()]
        self.deletedObjectsList = [obj for obj in objectList if obj.isDeleted()]

    def beginSync(self):
        if self.syncMode in [REFRESH_FROM_SERVER,
                             REFRESH_FROM_SERVER_BY_SERVER]:
            if self.objectList:
                if wx.MessageBox(_('The synchronization for source %s') % self.name + \
                                 _('will be a refresh from server. All local items will\n' \
                                   'be deleted. Do you wish to continue?'),
                                 _('Refresh from server'), wx.YES_NO) != wx.YES:
                    return 514 # Not sure of this
                self.objectList.clear()
                self.allObjectsList = []
                self.newObjectsList = []
                self.changedObjectsList = []
                self.deletedObjectsList = []
        elif self.syncMode in [REFRESH_FROM_CLIENT,
                               REFRESH_FROM_CLIENT_BY_SERVER]:
            if wx.MessageBox(_('The synchronization for source %s') % self.name + \
                             _('will be a refresh from client. All remote items will\n' \
                               'be deleted. Do you wish to continue?'),
                             _('Refresh from server'), wx.YES_NO) != wx.YES:
                return 514 # Not sure of this

    def endSync(self):
        pass

    def _getObject(self, key):
        """Returns the domain object with local ID 'key', or raise
        KeyError if not found."""

        for obj in self.allObjectsList:
            if obj.id() == key:
                return obj
        raise KeyError, 'No such object: %s' % key

    def _getItem(self, ls):
        """Returns a SyncItem instance representing the first domain
        object in the 'ls' sequence."""

        try:
            obj = ls.pop(0)
        except IndexError:
            return None

        item = SyncItem(obj.id())

        if obj.getStatus() == obj.STATUS_NONE:
            item.state = STATE_NONE
        elif obj.getStatus() == obj.STATUS_NEW:
            item.state = STATE_NEW
        elif obj.getStatus() == obj.STATUS_CHANGED:
            item.state = STATE_UPDATED
        elif obj.getStatus() == obj.STATUS_DELETED:
            item.state = STATE_DELETED

        self.updateItemProperties(item, obj)

        return item

    def updateItemProperties(self, item, obj):
        """Set item properties (data, dataType...) according to the
        domain object 'obj'. You must overload this in subclasses."""

        raise NotImplementedError

    def _parseObject(self, item):
        """Must return a new domain object from a SyncItem instance."""

        raise NotImplementedError

    # These two methods seem to be obsolete.

    def getFirstItemKey(self):
        return None

    def getNextItemKey(self):
        return None

    def getFirstItem(self):
        self.callback.pulse()

        self.allObjectsListCopy = self.allObjectsList[:]
        return self._getItem(self.allObjectsListCopy)

    def getNextItem(self):
        self.callback.pulse()

        return self._getItem(self.allObjectsListCopy)

    def getFirstNewItem(self):
        self.callback.pulse()

        self.newObjectsListCopy = self.newObjectsList[:]
        return self._getItem(self.newObjectsListCopy)

    def getNextNewItem(self):
        self.callback.pulse()

        return self._getItem(self.newObjectsListCopy)

    def getFirstUpdatedItem(self):
        self.callback.pulse()

        self.changedObjectsListCopy = self.changedObjectsList[:]
        return self._getItem(self.changedObjectsListCopy)

    def getNextUpdatedItem(self):
        self.callback.pulse()

        return self._getItem(self.changedObjectsListCopy)

    def getFirstDeletedItem(self):
        self.callback.pulse()

        self.deletedObjectsListCopy = self.deletedObjectsList[:]
        return self._getItem(self.deletedObjectsListCopy)

    def getNextDeletedItem(self):
        self.callback.pulse()

        return self._getItem(self.deletedObjectsListCopy)

    def addItem(self, item):
        self.callback.onAddItem()

        obj = self._parseObject(item)
        obj.cleanDirty()
        self.objectList.append(obj)
        item.key = obj.id().encode('UTF-8')

        return self.doAddItem(obj)

    def doAddItem(self, obj):
        """Called after a domain object has been added; use this to
        set up the object if it needs to."""

        return 201

    def updateItem(self, item):
        self.callback.onUpdateItem()

        obj = self._parseObject(item)

        try:
            local = self._getObject(item.key)
        except KeyError:
            return 404

        return self.doUpdateItem(obj, local)

    def doUpdateItem(self, obj, local):
        """Must update the 'local' domain object according to 'obj'
        (which is a 'temporary' domain object)"""
        local.cleanDirty()

    def deleteItem(self, item):
        self.callback.onDeleteItem()

        try:
            obj = self._getObject(item.key)
        except KeyError:
            return 211

        self.objectList.remove(obj)

        return 200

    def setItemStatus(self, key, status):
        self.callback.pulse()

        obj = self._getObject(key)

        if status in [200, 201, 211, 418]:
            # 200: Generic OK
            # 201: Added.
            # 211: Item not deleted (not found)
            # 418: Already exists.

            if obj.isDeleted():
                self.objectList.remove(obj)
            else:
                obj.cleanDirty()

            return 200

        print 'UNHANDLED ITEM STATUS %s %d' % (key, status)

        return 501



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import command
from taskcoachlib.domain import base, task, category, attachment
from taskcoachlib.gui import uicommand
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.pubsub import pub
import wx


class SearchableViewerMixin(object):
    ''' A viewer that is searchable. This is a mixin class. '''

    def isSearchable(self):
        return True
    
    def createFilter(self, presentation):
        presentation = super(SearchableViewerMixin, self).createFilter(presentation)
        return base.SearchFilter(presentation, **self.searchOptions())

    def searchOptions(self):
        searchString, matchCase, includeSubItems, searchDescription, regularExpression = self.getSearchFilter()
        return dict(searchString=searchString, 
                    matchCase=matchCase, 
                    includeSubItems=includeSubItems, 
                    searchDescription=searchDescription,
                    regularExpression=regularExpression,
                    treeMode=self.isTreeViewer())
    
    def setSearchFilter(self, searchString, matchCase=False, 
                        includeSubItems=False, searchDescription=False,
                        regularExpression=False):
        section = self.settingsSection()
        self.settings.set(section, 'searchfilterstring', searchString)
        self.settings.set(section, 'searchfiltermatchcase', str(matchCase))
        self.settings.set(section, 'searchfilterincludesubitems', str(includeSubItems))
        self.settings.set(section, 'searchdescription', str(searchDescription))
        self.settings.set(section, 'regularexpression', str(regularExpression))
        self.presentation().setSearchFilter(searchString, matchCase=matchCase, 
                                            includeSubItems=includeSubItems,
                                            searchDescription=searchDescription,
                                            regularExpression=regularExpression)
        
    def getSearchFilter(self):
        section = self.settingsSection()
        searchString = self.settings.get(section, 'searchfilterstring')
        matchCase = self.settings.getboolean(section, 'searchfiltermatchcase')
        includeSubItems = self.settings.getboolean(section, 'searchfilterincludesubitems')
        searchDescription = self.settings.getboolean(section, 'searchdescription')
        regularExpression = self.settings.getboolean(section, 'regularexpression')
        return searchString, matchCase, includeSubItems, searchDescription, regularExpression
    
    def createToolBarUICommands(self):
        ''' UI commands to put on the toolbar of this viewer. '''
        searchUICommand = uicommand.Search(viewer=self, settings=self.settings)
        return super(SearchableViewerMixin, self).createToolBarUICommands() + \
            (1, searchUICommand)
            

class FilterableViewerMixin(object):
    ''' A viewer that is filterable. This is a mixin class. '''
    
    def __init__(self, *args, **kwargs):
        self.__filterUICommands = None
        super(FilterableViewerMixin, self).__init__(*args, **kwargs)

    def isFilterable(self):
        return True

    def getFilterUICommands(self):
        if not self.__filterUICommands:
            self.__filterUICommands = self.createFilterUICommands()
        # Recreate the category filter commands every time because the category
        # filter menu depends on what categories there are
        return self.__filterUICommands[:2] + self.createCategoryFilterCommands() + self.__filterUICommands[2:]

    def createFilterUICommands(self):
        return [uicommand.ResetFilter(viewer=self), 
                uicommand.CategoryViewerFilterChoice(settings=self.settings),
                None]

    def createToolBarUICommands(self):
        clearUICommand = uicommand.ResetFilter(viewer=self)
        return super(FilterableViewerMixin, self).createToolBarUICommands() + \
            (clearUICommand,)

    def resetFilter(self):
        self.taskFile.categories().resetAllFilteredCategories()

    def hasFilter(self):
        return bool(self.taskFile.categories().filteredCategories())

    def createCategoryFilterCommands(self):
        categories = self.taskFile.categories()
        commands = [_('&Categories'), 
                uicommand.ResetCategoryFilter(categories=categories)]
        if categories:
            commands.append(None)
            commands.extend(self.createToggleCategoryFilterCommands(categories.rootItems()))
        return [tuple(commands)]
    
    def createToggleCategoryFilterCommands(self, categories):
        categories = list(categories)
        categories.sort(key=lambda category: category.subject())
        commands = [uicommand.ToggleCategoryFilter(category=eachCategory) for eachCategory in categories]
        categoriesWithChildren = [eachCategory for eachCategory in categories if eachCategory.children()]
        if categoriesWithChildren:
            commands.append(None)
            for eachCategory in categoriesWithChildren:
                subCommands = [_('%s (subcategories)') % eachCategory.subject()]
                subCommands.extend(self.createToggleCategoryFilterCommands(eachCategory.children()))
                commands.append(tuple(subCommands))
        return commands


class FilterableViewerForCategorizablesMixin(FilterableViewerMixin):
    def createFilter(self, items):
        items = super(FilterableViewerForCategorizablesMixin, self).createFilter(items)
        filterOnlyWhenAllCategoriesMatch = self.settings.getboolean('view', 
            'categoryfiltermatchall')
        return category.filter.CategoryFilter(items, 
            categories=self.taskFile.categories(), treeMode=self.isTreeViewer(), 
            filterOnlyWhenAllCategoriesMatch=filterOnlyWhenAllCategoriesMatch)


class FilterableViewerForTasksMixin(FilterableViewerForCategorizablesMixin):
    def createFilter(self, taskList):
        taskList = super(FilterableViewerForTasksMixin, self).createFilter(taskList)
        return task.filter.ViewFilter(taskList, treeMode=self.isTreeViewer(), 
                                      **self.viewFilterOptions())
        
    def viewFilterOptions(self):
        return dict(hideCompositeTasks=self.isHidingCompositeTasks(),
                    statusesToHide=self.hiddenTaskStatuses())
   
    def hideTaskStatus(self, status, hide=True):
        self.__setBooleanSetting('hide%stasks' % status, hide)
        self.presentation().hideTaskStatus(status, hide)

    def showOnlyTaskStatus(self, status):
        for taskStatus in task.Task.possibleStatuses():
            self.hideTaskStatus(taskStatus, hide=status != taskStatus)

    def isHidingTaskStatus(self, status):
        return self.__getBooleanSetting('hide%stasks' % status)
    
    def hiddenTaskStatuses(self):
        return [status for status in task.Task.possibleStatuses() if self.isHidingTaskStatus(status)]
    
    def hideCompositeTasks(self, hide=True):
        self.__setBooleanSetting('hidecompositetasks', hide)
        self.presentation().hideCompositeTasks(hide)
        
    def isHidingCompositeTasks(self):
        return self.__getBooleanSetting('hidecompositetasks')
    
    def resetFilter(self):
        super(FilterableViewerForTasksMixin, self).resetFilter()
        for status in task.Task.possibleStatuses():
            self.hideTaskStatus(status, False)
        if not self.isTreeViewer():
            # Only reset this filter when in list mode, since it only applies
            # to list mode
            self.hideCompositeTasks(False)

    def hasFilter(self):
        return super(FilterableViewerForTasksMixin, self).hasFilter() or \
            self.presentation().hasFilter()

    def createFilterUICommands(self):
        return super(FilterableViewerForTasksMixin, 
                     self).createFilterUICommands() + \
            [uicommand.ViewerHideTasks(taskStatus, viewer=self, settings=self.settings) for taskStatus in task.Task.possibleStatuses()] + \
            [uicommand.ViewerHideCompositeTasks(viewer=self)]
            
    def __getBooleanSetting(self, setting):
        return self.settings.getboolean(self.settingsSection(), setting)
    
    def __setBooleanSetting(self, setting, booleanValue):
        self.settings.setboolean(self.settingsSection(), setting, booleanValue)
        
                
class SortableViewerMixin(object):
    ''' A viewer that is sortable. This is a mixin class. '''

    def __init__(self, *args, **kwargs):
        self._sortUICommands = []
        super(SortableViewerMixin, self).__init__(*args, **kwargs)

    def isSortable(self):
        return True

    def registerPresentationObservers(self):
        super(SortableViewerMixin, self).registerPresentationObservers()
        pub.subscribe(self.onSortOrderChanged,
                      self.presentation().sortEventType())
        
    def detach(self):
        super(SortableViewerMixin, self).detach()
        pub.unsubscribe(self.onSortOrderChanged,
                        self.presentation().sortEventType())
        
    def onSortOrderChanged(self, sender):
        if sender == self.presentation():
            self.refresh()
            self.updateSelection(sendViewerStatusEvent=False)
            self.sendViewerStatusEvent()
        
    def createSorter(self, presentation):
        return self.SorterClass(presentation, **self.sorterOptions())

    def sorterOptions(self):
        return dict(sortBy=self.sortKey(),
                    sortCaseSensitive=self.isSortCaseSensitive())
        
    def sortBy(self, sortKey):
        self.presentation().sortBy(sortKey)
        self.settings.set(self.settingsSection(), 'sortby', str(self.presentation().sortKeys()))
        
    def isSortedBy(self, sortKey):
        sortKeys = self.presentation().sortKeys()
        return sortKeys and (sortKeys[0] == sortKey or sortKeys[0] == '-' + sortKey)

    def sortKey(self):
        return eval(self.settings.get(self.settingsSection(), 'sortby'))

    def isSortOrderAscending(self):
        sortKeys = self.presentation().sortKeys()
        return sortKeys and not sortKeys[0].startswith('-')

    def setSortOrderAscending(self, ascending=True):
        self.presentation().sortAscending(ascending)
        self.settings.set(self.settingsSection(), 'sortby', str(self.presentation().sortKeys()))

    def isSortCaseSensitive(self):
        return self.settings.getboolean(self.settingsSection(), 
            'sortcasesensitive')
        
    def setSortCaseSensitive(self, sortCaseSensitive=True):
        self.settings.set(self.settingsSection(), 'sortcasesensitive', 
            str(sortCaseSensitive))
        self.presentation().sortCaseSensitive(sortCaseSensitive)

    def getSortUICommands(self):
        if not self._sortUICommands:
            self.createSortUICommands()
        return self._sortUICommands

    def createSortUICommands(self):
        ''' (Re)Create the UICommands for sorting. These UICommands are put
            in the View->Sort menu and are used when the user clicks a column
            header. '''
        self._sortUICommands = self.createSortOrderUICommands()
        sortByCommands = self.createSortByUICommands()
        if sortByCommands:
            self._sortUICommands.append(None)  # Separator
            self._sortUICommands.extend(sortByCommands)
        
    def createSortOrderUICommands(self):
        ''' Create the UICommands for changing sort order, like ascending/
            descending, and match case. '''
        return [uicommand.ViewerSortOrderCommand(viewer=self),
                uicommand.ViewerSortCaseSensitive(viewer=self)]
        
    def createSortByUICommands(self):
        ''' Create the UICommands for changing what the items are sorted by,
            i.e. the columns. '''
        return [uicommand.ViewerSortByCommand(viewer=self, value='subject',
                    menuText=_('Sub&ject'),
                    helpText=self.sortBySubjectHelpText),
                uicommand.ViewerSortByCommand(viewer=self, value='description',
                    menuText=_('&Description'),
                    helpText=self.sortByDescriptionHelpText),
                uicommand.ViewerSortByCommand(viewer=self, 
                    value='creationDateTime', menuText=_('&Creation date'),
                    helpText=self.sortByCreationDateTimeHelpText),
                uicommand.ViewerSortByCommand(viewer=self,
                    value='modificationDateTime', 
                    menuText=_('&Modification date'),
                    helpText=self.sortByModificationDateTimeHelpText)]


class SortableViewerForEffortMixin(SortableViewerMixin):
    def createSortOrderUICommands(self):
        ''' Create the UICommands for changing sort order. The only
            option for efforts is ascending/descending at the moment. '''
        return [uicommand.ViewerSortOrderCommand(viewer=self)]

    def createSortByUICommands(self):
        ''' Create the UICommands for changing what the items are sorted by,
            i.e. the columns. Currently, effort is always sorted by period. '''
        return []

    def sortKey(self):
        ''' Efforts are always sorted by period at the moment. '''
        return ['-period']


class ManualOrderingMixin(object):
    def __init__(self, *args, **kwargs):
        if 'sort' not in self.viewerImages:
            self.viewerImages = self.viewerImages + ['sort']
        super(ManualOrderingMixin, self).__init__(*args, **kwargs)

    def createSortByUICommands(self):
        return [uicommand.ViewerSortByCommand(viewer=self, value='ordering',
                        menuText=_('&Manual ordering'),
                        helpText=self.sortByOrderingHelpText)] + super(ManualOrderingMixin, self).createSortByUICommands()

    def orderingImageIndices(self, item):
        index = self.imageIndex['sort']
        return {wx.TreeItemIcon_Normal: index,
                wx.TreeItemIcon_Expanded: index} 


class SortableViewerForCategoriesMixin(ManualOrderingMixin, SortableViewerMixin):
    sortBySubjectHelpText = _('Sort categories by subject')
    sortByDescriptionHelpText = _('Sort categories by description')
    sortByCreationDateTimeHelpText = _('Sort categories by creation date')
    sortByModificationDateTimeHelpText = _('Sort categories by last modification date')
    sortByOrderingHelpText = _('Sort categories manually')


class SortableViewerForCategorizablesMixin(SortableViewerMixin):
    ''' Mixin class to create uiCommands for sorting categorizables. '''

    def createSortByUICommands(self):
        commands = super(SortableViewerForCategorizablesMixin, self).createSortByUICommands()
        commands.append(uicommand.ViewerSortByCommand(viewer=self, 
            value='categories', menuText=_('&Category'),
            helpText=self.sortByCategoryHelpText))
        return commands


class SortableViewerForAttachmentsMixin(SortableViewerForCategorizablesMixin):
    sortBySubjectHelpText = _('Sort attachments by subject')
    sortByDescriptionHelpText = _('Sort attachments by description')
    sortByCategoryHelpText = _('Sort attachments by category')
    sortByCreationDateTimeHelpText = _('Sort attachments by creation date')
    sortByModificationDateTimeHelpText = _('Sort attachments by last modification date')


class SortableViewerForNotesMixin(ManualOrderingMixin, SortableViewerForCategorizablesMixin):
    sortBySubjectHelpText = _('Sort notes by subject')
    sortByDescriptionHelpText = _('Sort notes by description')
    sortByCategoryHelpText = _('Sort notes by category')
    sortByCreationDateTimeHelpText = _('Sort notes by creation date')
    sortByModificationDateTimeHelpText = _('Sort notes by last modification date')
    sortByOrderingHelpText = _('Sort notes manually')


class SortableViewerForTasksMixin(ManualOrderingMixin, SortableViewerForCategorizablesMixin):
    SorterClass = task.sorter.Sorter
    sortBySubjectHelpText = _('Sort tasks by subject')
    sortByDescriptionHelpText = _('Sort tasks by description')
    sortByCategoryHelpText = _('Sort tasks by category')
    sortByCreationDateTimeHelpText = _('Sort tasks by creation date')
    sortByModificationDateTimeHelpText = _('Sort tasks by last modification date')
    sortByOrderingHelpText = _('Sort tasks manually')

    def __init__(self, *args, **kwargs):
        self.__sortKeyUnchangedCount = 0
        super(SortableViewerForTasksMixin, self).__init__(*args, **kwargs)
    
    def sortBy(self, sortKey):
        # If the user clicks the same column for the third time, toggle
        # the SortyByTaskStatusFirst setting:
        if self.isSortedBy(sortKey):
            self.__sortKeyUnchangedCount += 1
        else:
            self.__sortKeyUnchangedCount = 0
        if self.__sortKeyUnchangedCount > 1:
            self.setSortByTaskStatusFirst(not self.isSortByTaskStatusFirst())
            self.__sortKeyUnchangedCount = 0
        super(SortableViewerForTasksMixin, self).sortBy(sortKey)

    def isSortByTaskStatusFirst(self):
        return self.settings.getboolean(self.settingsSection(),
            'sortbystatusfirst')
        
    def setSortByTaskStatusFirst(self, sortByTaskStatusFirst):
        self.settings.set(self.settingsSection(), 'sortbystatusfirst',
            str(sortByTaskStatusFirst))
        self.presentation().sortByTaskStatusFirst(sortByTaskStatusFirst)
        
    def sorterOptions(self):
        options = super(SortableViewerForTasksMixin, self).sorterOptions()
        options.update(treeMode=self.isTreeViewer(), 
            sortByTaskStatusFirst=self.isSortByTaskStatusFirst())
        return options

    def createSortOrderUICommands(self):
        commands = super(SortableViewerForTasksMixin, self).createSortOrderUICommands()
        commands.append(uicommand.ViewerSortByTaskStatusFirst(viewer=self))
        return commands
    
    def createSortByUICommands(self):
        commands = super(SortableViewerForTasksMixin, self).createSortByUICommands()
        dependsOnEffortFeature = ['budget', 'timeSpent', 'budgetLeft',  
                                  'hourlyFee', 'fixedFee', 'revenue']
        for menuText, helpText, value in [\
            (_('&Planned start date'), _('Sort tasks by planned start date'), 'plannedStartDateTime'),
            (_('&Due date'), _('Sort tasks by due date'), 'dueDateTime'),
            (_('&Completion date'), _('Sort tasks by completion date'), 'completionDateTime'),
            (_('&Prerequisites'), _('Sort tasks by prerequisite tasks'), 'prerequisites'),
            (_('&Dependents'), _('Sort tasks by dependent tasks'), 'dependencies'),
            (_('&Time left'), _('Sort tasks by time left'), 'timeLeft'),
            (_('&Percentage complete'), _('Sort tasks by percentage complete'), 'percentageComplete'),
            (_('&Recurrence'), _('Sort tasks by recurrence'), 'recurrence'),
            (_('&Budget'), _('Sort tasks by budget'), 'budget'),
            (_('&Time spent'), _('Sort tasks by time spent'), 'timeSpent'),
            (_('Budget &left'), _('Sort tasks by budget left'), 'budgetLeft'),
            (_('&Priority'), _('Sort tasks by priority'), 'priority'),
            (_('&Hourly fee'), _('Sort tasks by hourly fee'), 'hourlyFee'),
            (_('&Fixed fee'), _('Sort tasks by fixed fee'), 'fixedFee'),
            (_('&Revenue'), _('Sort tasks by revenue'), 'revenue'),
            (_('&Reminder'), _('Sort tasks by reminder date and time'), 'reminder')]:
            if value not in dependsOnEffortFeature or (value in dependsOnEffortFeature):
                commands.append(uicommand.ViewerSortByCommand(\
                    viewer=self, value=value, menuText=menuText, helpText=helpText))
        return commands
    

class AttachmentDropTargetMixin(object):
    ''' Mixin class for viewers that are drop targets for attachments. '''

    def widgetCreationKeywordArguments(self):
        kwargs = super(AttachmentDropTargetMixin, self).widgetCreationKeywordArguments()
        kwargs['onDropURL'] = self.onDropURL
        kwargs['onDropFiles'] = self.onDropFiles
        kwargs['onDropMail'] = self.onDropMail
        return kwargs
        
    def _addAttachments(self, attachments, item, **itemDialogKwargs):
        ''' Add attachments. If item refers to an existing domain object, 
            add the attachments to that object. If item is None, use the 
            newItemDialog to create a new domain object and add the attachments
            to that new object. '''
        if item is None:
            itemDialogKwargs['subject'] = attachments[0].subject()
            if self.settings.get('view', 'defaultplannedstartdatetime').startswith('preset'):
                itemDialogKwargs['plannedStartDateTime'] = task.Task.suggestedPlannedStartDateTime()
            if self.settings.get('view', 'defaultduedatetime').startswith('preset'):
                itemDialogKwargs['dueDateTime'] = task.Task.suggestedDueDateTime()
            if self.settings.get('view', 'defaultactualstartdatetime').startswith('preset'):
                itemDialogKwargs['actualStartDateTime'] = task.Task.suggestedActualStartDateTime()
            if self.settings.get('view', 'defaultcompletiondatetime').startswith('preset'):
                itemDialogKwargs['completionDateTime'] = task.Task.suggestedCompletionDateTime()
            if self.settings.get('view', 'defaultreminderdatetime').startswith('preset'):
                itemDialogKwargs['reminder'] = task.Task.suggestedReminderDateTime()
            newItemDialog = self.newItemDialog(bitmap='new',
                attachments=attachments, **itemDialogKwargs)
            newItemDialog.Show()
        else:
            addAttachment = command.AddAttachmentCommand(self.presentation(),
                [item], attachments=attachments)
            addAttachment.do()

    def onDropURL(self, item, url, **kwargs):
        ''' This method is called by the widget when a URL is dropped on an 
            item. '''
        attachments = [attachment.URIAttachment(url)]
        self._addAttachments(attachments, item, **kwargs)

    def onDropFiles(self, item, filenames, **kwargs):
        ''' This method is called by the widget when one or more files
            are dropped on an item. '''
        attachmentBase = self.settings.get('file', 'attachmentbase')
        if attachmentBase:
            filenames = [attachment.getRelativePath(filename, attachmentBase) \
                         for filename in filenames]
        attachments = [attachment.FileAttachment(filename) for filename in filenames]
        self._addAttachments(attachments, item, **kwargs)

    def onDropMail(self, item, mail, **kwargs):
        ''' This method is called by the widget when a mail message is dropped
            on an item. '''
        att = attachment.MailAttachment(mail)
        subject, content = att.read()
        self._addAttachments([att], item, subject=subject, description=content, 
                             **kwargs)


class NoteColumnMixin(object):
    def noteImageIndices(self, item):
        index = self.imageIndex['note_icon'] if item.notes() else -1
        return {wx.TreeItemIcon_Normal: index}
    

class AttachmentColumnMixin(object):    
    def attachmentImageIndices(self, item):  # pylint: disable=W0613
        index = self.imageIndex['paperclip_icon'] if item.attachments() else -1
        return {wx.TreeItemIcon_Normal: index}



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx, sys, platform

# This module is meant to be imported like this: 
#   from taskcoachlib import operating_system
# so that the function calls read: 
#   operating_system.isWindows(), operating_system.isMac(), etc.

def isMac():
    return isPlatform('MAC')


def isWindows():
    return isPlatform('MSW')


def isGTK():
    return isPlatform('GTK')


def isPlatform(threeLetterPlatformAbbreviation, wxPlatform=wx.Platform):
    return '__WX%s__'%threeLetterPlatformAbbreviation == wxPlatform


def isWindows7_OrNewer(): # pragma: no cover
    if isWindows(): 
        major, minor = sys.getwindowsversion()[:2] # pylint: disable=E1101
        return (major, minor) >= (6, 1)
    else:
        return False


def _platformVersion():
    return tuple(map(int, platform.release().split('.')))

def isMacOsXLion_OrNewer(): # pragma: no cover
    if isMac():
        return _platformVersion() >= (11, 1)
    else:
        return False

def isMacOsXTiger_OrOlder(): # pragma no cover
    if isMac():
        return _platformVersion() <= (8, 11, 1) # Darwin release number for Tiger
    else:
        return False

def isMacOsXMountainLion_OrNewer(): # pragma no cover
    if isMac():
        return _platformVersion() >= (12,)
    else:
        return False

def isMacOsXMavericks_OrNewer(): # pragma no cover
    if isMac():
        return _platformVersion() >= (13,)
    else:
        return False


def defaultEncodingName():
    return wx.Locale.GetSystemEncodingName() or 'utf-8'

def decodeSystemString(s):
    if isinstance(s, unicode):
        return s
    encoding = defaultEncodingName()
    # Python does not define the windows_XXX aliases for every code page...
    if encoding.startswith('windows-'):
        encoding = 'cp' + encoding[8:]
    if not encoding:
        encoding = 'utf-8'
    return s.decode(encoding, 'ignore')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.gui import dialog
from taskcoachlib import config
from taskcoachlib.domain import task, effort


class DummyEvent(object):
    def Skip(self):
        pass


class ReminderDialogTest(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        task.Task.settings = self.settings
        self.aTask = task.Task('subject')
        self.taskList = task.TaskList([self.aTask])
        self.effortList = effort.EffortList(self.taskList)
        
    def createReminderDialog(self):
        return dialog.reminder.ReminderDialog(self.aTask, 
            self.taskList, self.effortList, self.settings, None)

    def testRememberZeroSnoozeTime(self):
        reminderDialog = self.createReminderDialog()
        reminderDialog.snoozeOptions.SetSelection(0)
        reminderDialog.onClose(DummyEvent())        
        self.assertEqual(0, self.settings.getint('view', 'defaultsnoozetime'))

    def testRememberSnoozeTime(self):
        reminderDialog = self.createReminderDialog()
        reminderDialog.snoozeOptions.SetSelection(2)
        reminderDialog.onClose(DummyEvent())        
        self.assertEqual(10, self.settings.getint('view', 'defaultsnoozetime'))

    def testUseDefaultSnoozeTime(self):
        self.settings.set('view', 'defaultsnoozetime', '15')
        reminderDialog = self.createReminderDialog()
        self.assertEqual('15 minutes', 
                         reminderDialog.snoozeOptions.GetStringSelection())
        
    def testDontUseDefaultSnoozeTimeWhenItsNotInTheListOfOptions(self):
        self.settings.set('view', 'defaultsnoozetime', '17')
        reminderDialog = self.createReminderDialog()
        self.assertEqual('5 minutes', 
                         reminderDialog.snoozeOptions.GetStringSelection())
        
    def testRememberReminderReplaceDefaultSnoozeTime(self):
        reminderDialog = self.createReminderDialog()
        reminderDialog.replaceDefaultSnoozeTime.SetValue(False)
        reminderDialog.onClose(DummyEvent())        
        self.assertEqual(False, self.settings.getboolean('view', 
                                'replacedefaultsnoozetime'))
        
    def testUseReminderReplaceDefaultSnoozeTime(self):
        self.settings.setboolean('view', 'replacedefaultsnoozetime', False)
        reminderDialog = self.createReminderDialog()
        self.assertEqual(False, 
                         reminderDialog.replaceDefaultSnoozeTime.GetValue())
        


'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.thirdparty.wxScheduler import wxScheduler, wxSchedule, \
    EVT_SCHEDULE_ACTIVATED, EVT_SCHEDULE_RIGHT_CLICK, \
    EVT_SCHEDULE_DCLICK, EVT_PERIODWIDTH_CHANGED, wxReportScheduler, wxTimeFormat
from taskcoachlib.thirdparty.wxScheduler.wxSchedulerConstants import wxSCHEDULER_WEEKSTART_MONDAY,\
    wxSCHEDULER_WEEKSTART_SUNDAY
from taskcoachlib.domain import date
from taskcoachlib.widgets import draganddrop
from taskcoachlib import command, render
import tooltip


class _CalendarContent(tooltip.ToolTipMixin, wxScheduler):
    def __init__(self, parent, taskList, iconProvider, onSelect, onEdit,
                 onCreate, onChangeConfig, popupMenu, *args, **kwargs):
        self.getItemTooltipData = parent.getItemTooltipData

        self.__onDropURLCallback = kwargs.pop('onDropURL', None)
        self.__onDropFilesCallback = kwargs.pop('onDropFiles', None)
        self.__onDropMailCallback = kwargs.pop('onDropMail', None)

        self.dropTarget = draganddrop.DropTarget(self.OnDropURL,
                                                 self.OnDropFiles,
                                                 self.OnDropMail)

        super(_CalendarContent, self).__init__(parent, wx.ID_ANY, 
                                               *args, **kwargs)

        self.SetDropTarget(self.dropTarget)

        self.selectCommand = onSelect
        self.iconProvider = iconProvider
        self.editCommand = onEdit
        self.createCommand = onCreate
        self.changeConfigCb = onChangeConfig
        self.popupMenu = popupMenu

        self.__tip = tooltip.SimpleToolTip(self)
        self.__selection = []

        self.__showNoPlannedStartDate = False
        self.__showNoDueDate = False
        self.__showUnplanned = False

        self.SetShowWorkHour(False)
        self.SetResizable(True)

        self.taskList = taskList
        self.RefreshAllItems(0)

        EVT_SCHEDULE_ACTIVATED(self, self.OnActivation)
        EVT_SCHEDULE_RIGHT_CLICK(self, self.OnPopup)
        EVT_SCHEDULE_DCLICK(self, self.OnEdit)
        EVT_PERIODWIDTH_CHANGED(self, self.OnChangeConfig)

        wxTimeFormat.SetFormatFunction(self.__formatTime)

    @staticmethod
    def __formatTime(dateTime, includeMinutes=False):
        return render.time(TaskSchedule.tcDateTime(dateTime), 
                           minutes=includeMinutes)

    def _handleDrop(self, x, y, droppedObject, cb):
        if cb is not None:
            _, _, item = self._findSchedule(wx.Point(x, y))

            if item is not None:
                if isinstance(item, TaskSchedule):
                    cb(item.task, droppedObject)
                else:
                    datetime = date.DateTime(item.GetYear(), 
                                             item.GetMonth() + 1, 
                                             item.GetDay())
                    cb(None, droppedObject, plannedStartDateTime=datetime, 
                       dueDateTime=datetime.endOfDay())

    def GetPrintout(self, settings):
        return wxReportScheduler(self.GetViewType(),
                                 self.GetStyle(),
                                 self.GetDrawer(),
                                 self.GetDate(),
                                 self.GetWeekStart(),
                                 self.GetPeriodCount(),
                                 self.GetSchedules())

    def OnDropURL(self, x, y, url):
        self._handleDrop(x, y, url, self.__onDropURLCallback)

    def OnDropFiles(self, x, y, filenames):
        self._handleDrop(x, y, filenames, self.__onDropFilesCallback)

    def OnDropMail(self, x, y, mail):
        self._handleDrop(x, y, mail, self.__onDropMailCallback)

    def SetShowNoStartDate(self, doShow):
        self.__showNoPlannedStartDate = doShow
        self.RefreshAllItems(0)

    def SetShowNoDueDate(self, doShow):
        self.__showNoDueDate = doShow
        self.RefreshAllItems(0)

    def SetShowUnplanned(self, doShow):
        self.__showUnplanned = doShow
        self.RefreshAllItems(0)

    def OnChangeConfig(self, event):  # pylint: disable=W0613
        self.changeConfigCb()

    def Select(self, schedule=None):
        if self.__selection:
            self.taskMap[self.__selection[0].id()].SetSelected(False)

        if schedule is None:
            self.__selection = []
        else:
            self.__selection = [schedule.task]
            schedule.SetSelected(True)

        wx.CallAfter(self.selectCommand)

    def SelectTask(self, task):
        if task.id() in self.taskMap:
            self.Select(self.taskMap[task.id()])

    def OnActivation(self, event):
        self.SetFocus()
        self.Select(event.schedule)

    def OnPopup(self, event):
        self.OnActivation(event)
        wx.CallAfter(self.PopupMenu, self.popupMenu)

    def OnEdit(self, event):
        if event.schedule is None:
            if event.date is not None:
                self.createCommand(date.DateTime(event.date.GetYear(),
                                                 event.date.GetMonth() + 1,
                                                 event.date.GetDay(),
                                                 event.date.GetHour(),
                                                 event.date.GetMinute(),
                                                 event.date.GetSecond()))
        else:
            self.editCommand(event.schedule.task)

    def RefreshAllItems(self, count):  # pylint: disable=W0613
        x, y = self.GetViewStart()
        selectionId = None
        if self.__selection:
            selectionId = self.__selection[0].id()
        self.__selection = []

        self.DeleteAll()

        schedules = []
        self.taskMap = {}  # pylint: disable=W0201
        maxDateTime = date.DateTime()

        for task in self.taskList:
            if not task.isDeleted():
                if task.plannedStartDateTime() == maxDateTime or not task.completed():
                    if task.plannedStartDateTime() == maxDateTime and not self.__showNoPlannedStartDate:
                        continue

                    if task.dueDateTime() == maxDateTime and not self.__showNoDueDate:
                        continue

                    if not self.__showUnplanned:
                        if task.plannedStartDateTime() == maxDateTime and task.dueDateTime() == maxDateTime:
                            continue

                schedule = TaskSchedule(task, self.iconProvider)
                schedules.append(schedule)
                self.taskMap[task.id()] = schedule

                if task.id() == selectionId:
                    self.__selection = [task]
                    schedule.SetSelected(True)

        self.Add(schedules)
        wx.CallAfter(self.selectCommand)
        self.Scroll(x, y)

    def RefreshItems(self, *args):
        selectionId = None
        if self.__selection:
            selectionId = self.__selection[0].id()
        self.__selection = []

        for task in args:
            doShow = True

            if task.plannedStartDateTime() == date.DateTime() and task.dueDateTime() == date.DateTime() and not self.__showUnplanned:
                doShow = False

            if task.plannedStartDateTime() == date.DateTime() and not self.__showNoPlannedStartDate:
                doShow = False

            if task.dueDateTime() == date.DateTime() and not self.__showNoDueDate:
                doShow = False

            # Special case

            if task.isDeleted():
                doShow = False
            elif task.plannedStartDateTime() != date.DateTime() and task.completed():
                doShow = True

            if doShow:
                if self.taskMap.has_key(task.id()):
                    schedule = self.taskMap[task.id()]
                    schedule.update()
                else:
                    schedule = TaskSchedule(task, self.iconProvider)
                    self.taskMap[task.id()] = schedule
                    self.Add([schedule])

                if task.id() == selectionId:
                    self.__selection = [task]
                    schedule.SetSelected(True)
            else:
                if self.taskMap.has_key(task.id()):
                    self.Delete(self.taskMap[task.id()])
                    del self.taskMap[task.id()]
                    if self.__selection and self.__selection[0].id() == task.id():
                        self.__selection = []
                        wx.CallAfter(self.selectCommand)

    def GetItemCount(self):
        return len(self.GetSchedules())
        
    def OnBeforeShowToolTip(self, x, y):
        originX, originY = self.GetViewStart()
        unitX, unitY = self.GetScrollPixelsPerUnit()

        try:
            _, _, schedule = self._findSchedule(wx.Point(x + originX * unitX, 
                                                         y + originY * unitY))
        except TypeError:
            return

        if schedule and isinstance(schedule, TaskSchedule):
            item = schedule.task

            tooltipData = self.getItemTooltipData(item)
            doShow = any(data[1] for data in tooltipData)
            if doShow:
                self.__tip.SetData(tooltipData)
                return self.__tip
            else:
                return None

    def GetMainWindow(self):
        return self
    
    MainWindow = property(GetMainWindow)

    def curselection(self):
        return self.__selection


class Calendar(wx.Panel):
    def __init__(self, parent, taskList, iconProvider, onSelect, onEdit,
                 onCreate, popupMenu, *args, **kwargs):
        self.getItemTooltipData = parent.getItemTooltipData

        super(Calendar, self).__init__(parent)

        self._headers = wx.Panel(self)
        self._content = _CalendarContent(self, taskList, iconProvider, 
                                         onSelect, onEdit, onCreate, popupMenu, 
                                         *args, **kwargs)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self._headers, 0, wx.EXPAND)
        sizer.Add(self._content, 1, wx.EXPAND)
        self.SetSizer(sizer)

        # Must wx.CallAfter because SetDrawerClass is called this way.
        wx.CallAfter(self._content.SetHeaderPanel, self._headers)

    def Draw(self, dc):
        self._content.Draw(dc)

    def SetShowNoStartDate(self, doShow):
        self._content.SetShowNoStartDate(doShow)

    def SetShowNoDueDate(self, doShow):
        self._content.SetShowNoDueDate(doShow)

    def SetShowUnplanned(self, doShow):
        self._content.SetShowUnplanned(doShow)
        
    def SetWeekStartMonday(self):
        self._content.SetWeekStart(wxSCHEDULER_WEEKSTART_MONDAY)
        
    def SetWeekStartSunday(self):
        self._content.SetWeekStart(wxSCHEDULER_WEEKSTART_SUNDAY)

    def RefreshAllItems(self, count):
        self._content.RefreshAllItems(count)

    def RefreshItems(self, *args):
        self._content.RefreshItems(*args)

    def GetItemCount(self):
        return self._content.GetItemCount()

    def curselection(self):
        return self._content.curselection()

    def isAnyItemCollapsable(self):
        return False

    def select(self, tasks):
        if len(tasks) == 1:
            self._content.SelectTask(tasks[0])
        else:
            self._content.Select(None)

    def __getattr__(self, name):
        return getattr(self._content, name)


class TaskSchedule(wxSchedule):
    def __init__(self, task, iconProvider):
        super(TaskSchedule, self).__init__()

        self.__selected = False

        self.clientdata = task
        self.iconProvider = iconProvider
        self.update()

    def SetSelected(self, selected):
        self.Freeze()
        try:
            self.__selected = selected
            if selected:
                self.color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
                # On MS Windows, the selection background is very dark. If
                # the foreground color is too dark, invert it.
                color = self.task.foregroundColor(True) or (0, 0, 0)
                if len(color) == 3:
                    r, g, b = color
                else:
                    r, g, b, a = color
                if r + g + b < 128 * 3:
                    self.foreground = wx.Colour(255 - r, 255 - g, 255 - b)
            else:
                self.color = wx.Colour(*(self.task.backgroundColor(True) or (255, 255, 255)))
                self.foreground = wx.Colour(*(self.task.foregroundColor(True) or (0, 0, 0)))
        finally:
            self.Thaw()

    def SetStart(self, start):
        command.EditPlannedStartDateTimeCommand(items=[self.task], newValue=self.tcDateTime(start)).do()

    def SetEnd(self, end):
        if self.task.completed():
            command.EditCompletionDateTimeCommand(items=[self.task], 
                                                  newValue=self.tcDateTime(end)).do()
        else:
            command.EditDueDateTimeCommand(items=[self.task], 
                                           newValue=self.tcDateTime(end)).do()

    def Offset(self, ts):
        kwargs = dict()
        if self.task.plannedStartDateTime() != date.DateTime():
            start = self.GetStart()
            start.AddTS(ts)
            command.EditPlannedStartDateTimeCommand(items=[self.task], 
                                                    newValue=self.tcDateTime(start)).do()
        if self.task.completed():
            end = self.GetEnd()
            end.AddTS(ts)
            command.EditCompletionDateTimeCommand(items=[self.task], 
                                                  newValue=self.tcDateTime(end)).do()
        elif self.task.dueDateTime() != date.DateTime():
            end = self.GetEnd()
            end.AddTS(ts)
            command.EditDueDateTimeCommand(items=[self.task], 
                                           newValue=self.tcDateTime(end)).do()

    @property
    def task(self):
        return self.clientdata

    def update(self):
        self.Freeze()
        try:
            self.description = self.task.subject()

            self.start = self.wxDateTime(self.task.plannedStartDateTime(),
                                         self.tupleFromDateTime(date.Now().startOfDay()))
            end = self.task.completionDateTime() if self.task.completed() else self.task.dueDateTime()
            self.end = self.wxDateTime(end,
                                       self.tupleFromDateTime(date.Now().endOfDay()))
            
            if self.task.completed():
                self.done = True

            self.color = wx.Colour(*(self.task.backgroundColor(True) or (255, 255, 255)))
            self.foreground = wx.Colour(*(self.task.foregroundColor(True) or (0, 0, 0)))
            self.font = self.task.font(True)

            self.icons = [self.iconProvider(self.task, False)]
            if self.task.attachments():
                self.icons.append('paperclip_icon')
            if self.task.notes():
                self.icons.append('note_icon')

            if self.task.percentageComplete(recursive=True):
                # If 0, just let the default None value so the progress bar isn't drawn
                # at all
                self.complete = 1.0 * self.task.percentageComplete(recursive=True) / 100
            else:
                self.complete = None
        finally:
            self.Thaw()

    @staticmethod
    def tupleFromDateTime(dateTime):
        return (dateTime.day,
                dateTime.month - 1,
                dateTime.year,
                dateTime.hour,
                dateTime.minute,
                dateTime.second)

    @staticmethod
    def wxDateTime(dateTime, default):
        args = default if dateTime == date.DateTime() else \
            (dateTime.day, dateTime.month - 1, dateTime.year,
             dateTime.hour, dateTime.minute, dateTime.second)
        return wx.DateTimeFromDMY(*args) # pylint: disable=W0142

    @staticmethod
    def tcDateTime(dateTime):
        return date.DateTime(dateTime.GetYear(),
                             dateTime.GetMonth() + 1,
                             dateTime.GetDay(),
                             dateTime.GetHour(),
                             dateTime.GetMinute(),
                             dateTime.GetSecond())




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import config
from taskcoachlib.domain import task, effort, date

 
class TaskListTest(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.taskList = task.TaskList()
        year = date.Now().year
        self.task1 = task.Task(dueDateTime=date.DateTime(year+1,1,1))
        self.task2 = task.Task(dueDateTime=date.DateTime(year+2,1,1))
        self.task3 = task.Task()
        
    def nrStatus(self, status):
        return self.taskList.nrOfTasksPerStatus()[status]
    
    def testNrOfTasksPerStatusOfAnEmptyTaskList(self):
        counts = self.taskList.nrOfTasksPerStatus()
        for status in task.Task.possibleStatuses():     
            self.assertEqual(0, counts[status])
            
    def testNrCompleted(self):
        self.assertEqual(0, self.nrStatus(task.status.completed))
        self.taskList.append(self.task1)
        self.assertEqual(0, self.nrStatus(task.status.completed))
        self.task1.setCompletionDateTime()
        self.assertEqual(1, self.nrStatus(task.status.completed))
    
    def testNrOverdue(self):
        self.assertEqual(0, self.nrStatus(task.status.overdue))
        self.taskList.append(self.task1)
        self.assertEqual(0, self.nrStatus(task.status.overdue))
        self.task1.setDueDateTime(date.DateTime(1990, 1, 1))
        self.assertEqual(1, self.nrStatus(task.status.overdue))

    def testNrDueSoon(self):
        self.assertEqual(0, self.nrStatus(task.status.duesoon))
        self.taskList.append(task.Task(dueDateTime=date.Now() + date.ONE_HOUR))
        self.assertEqual(1, self.nrStatus(task.status.duesoon))
        
    def testNrBeingTracked(self):
        self.assertEqual(0, self.taskList.nrBeingTracked())
        activeTask = task.Task()
        activeTask.addEffort(effort.Effort(activeTask))
        self.taskList.append(activeTask)
        self.assertEqual(1, self.taskList.nrBeingTracked())
        
    def testOriginalLength(self):
        self.assertEqual(0, self.taskList.originalLength())

    def testMinPriority_EmptyTaskList(self):
        self.assertEqual(0, self.taskList.minPriority())
        
    def testMinPriority_OneTaskWithDefaultPriority(self):
        self.taskList.append(self.task1)
        self.assertEqual(self.task1.priority(), self.taskList.minPriority())
        
    def testMinPriority_OneTaskWithNonDefaultPriority(self):
        self.taskList.append(task.Task(priority=-5))
        self.assertEqual(-5, self.taskList.minPriority())
        
    def testMinPriority_TwoTasks(self):
        self.taskList.extend([task.Task(priority=3), task.Task(priority=5)])
        self.assertEqual(3, self.taskList.minPriority())
        
    def testMaxPriority_EmptyTaskList(self):
        self.assertEqual(0, self.taskList.maxPriority())
        
    def testMaxPriority_OneTaskWithDefaultPriority(self):
        self.taskList.append(self.task1)
        self.assertEqual(self.task1.priority(), self.taskList.maxPriority())
        
    def testMaxPriority_OneTaskWithNonDefaultPriority(self):
        self.taskList.append(task.Task(priority=-5))
        self.assertEqual(-5, self.taskList.maxPriority())
        
    def testMaxPriority_TwoTasks(self):
        self.taskList.extend([task.Task(priority=3), task.Task(priority=5)])
        self.assertEqual(5, self.taskList.maxPriority())



'''
Task Coach - Your friendly task manager
Copyright (C) 2012 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

# Not using agw.balloontip because it doesn't position properly and
# lacks events

import wx
from wx.lib.embeddedimage import PyEmbeddedImage


class BalloonTip(wx.Frame):
    ARROWSIZE = 16
    MAXWIDTH = 300

    def __init__(self, parent, target, message=None, title=None, bitmap=None, getRect=None):
        """Baloon tip."""

        super(BalloonTip, self).__init__(parent,
            style=wx.NO_BORDER|wx.FRAME_FLOAT_ON_PARENT|wx.FRAME_NO_TASKBAR|wx.FRAME_SHAPED|wx.POPUP_WINDOW)

        wheat = wx.ColourDatabase().Find('WHEAT')
        self.SetBackgroundColour(wheat)

        self._target = target
        self._getRect = getRect
        self._interior = wx.Panel(self)
        self._interior.Bind(wx.EVT_LEFT_DOWN, self.DoClose)
        self._interior.SetBackgroundColour(wheat)
        vsizer = wx.BoxSizer(wx.VERTICAL)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        if bitmap is not None:
            hsizer.Add(wx.StaticBitmap(self._interior, wx.ID_ANY, bitmap), 0, wx.ALIGN_CENTRE|wx.ALL, 3)
        if title is not None:
            titleCtrl = wx.StaticText(self._interior, wx.ID_ANY, title)
            hsizer.Add(titleCtrl, 1, wx.ALL|wx.ALIGN_CENTRE, 3)
            titleCtrl.Bind(wx.EVT_LEFT_DOWN, self.DoClose)
        vsizer.Add(hsizer, 0, wx.EXPAND)
        if message is not None:
            msg = wx.StaticText(self._interior, wx.ID_ANY, message)
            msg.Wrap(self.MAXWIDTH)
            vsizer.Add(msg, 1, wx.EXPAND|wx.ALL, 3)
            msg.Bind(wx.EVT_LEFT_DOWN, self.DoClose)

        self._interior.SetSizer(vsizer)
        self._interior.Fit()

        class Sizer(wx.PySizer):
            def __init__(self, interior, direction, offset):
                self._interior = interior
                self._direction = direction
                self._offset = offset
                super(Sizer, self).__init__()

            def SetDirection(self, direction):
                self._direction = direction

            def CalcMin(self):
                w, h = self._interior.GetClientSize()
                return wx.Size(w, h + self._offset)

            def RecalcSizes(self):
                if self._direction == 'bottom':
                    self._interior.SetPosition((0, 0))
                else:
                    self._interior.SetPosition((0, self._offset))

        self._sizer = Sizer(self._interior, 'bottom', self.ARROWSIZE)
        self.SetSizer(self._sizer)
        self.Position()
        self.Show()

        wx.GetTopLevelParent(target).Bind(wx.EVT_SIZE, self._OnDim)
        wx.GetTopLevelParent(target).Bind(wx.EVT_MOVE, self._OnDim)

    def _Unbind(self):
        wx.GetTopLevelParent(self._target).Unbind(wx.EVT_SIZE)
        wx.GetTopLevelParent(self._target).Unbind(wx.EVT_MOVE)

    def _OnDim(self, event):
        wx.CallAfter(self.Position)
        event.Skip()

    def DoClose(self, event, unbind=True):
        if unbind:
            self._Unbind()
        self.Close()

    def Position(self):
        w, h = self._interior.GetClientSizeTuple()
        h += self.ARROWSIZE
        if self._getRect is None:
            tw, th = self._target.GetSizeTuple()
            tx, ty = 0, 0
        else:
            tx, ty, tw, th = self._getRect()
        tx, ty = self._target.ClientToScreen(wx.Point(tx, ty))
        dpyIndex = max(0, wx.Display.GetFromPoint(wx.Point(tx, ty)) or 0)
        rect = wx.Display(dpyIndex).GetClientArea()

        x = max(rect.GetLeft(), min(rect.GetRight() - w, int(tx + tw / 2 - w / 2)))
        y = ty - h
        direction = 'bottom'
        if y < rect.GetTop():
            y = ty + th
            direction = 'top'

        mask = wx.EmptyBitmap(w, h)
        memDC = wx.MemoryDC()
        memDC.SelectObject(mask)
        try:
            memDC.SetBrush(wx.BLACK_BRUSH)
            memDC.SetPen(wx.BLACK_PEN)
            memDC.DrawRectangle(0, 0, w, h)

            memDC.SetBrush(wx.WHITE_BRUSH)
            memDC.SetPen(wx.WHITE_PEN)
            if direction == 'bottom':
                memDC.DrawPolygon([(0, 0),
                                   (w, 0),
                                   (w, h - self.ARROWSIZE),
                                   (tx + int(tw / 2) - x + int(self.ARROWSIZE / 2), h - self.ARROWSIZE),
                                   (tx + int(tw / 2) - x, h),
                                   (tx + int(tw / 2) - x - int(self.ARROWSIZE / 2), h - self.ARROWSIZE),
                                   (0, h - self.ARROWSIZE)])
            else:
                memDC.DrawPolygon([(0, self.ARROWSIZE),
                                   (tx + int(tw / 2) - x - int(self.ARROWSIZE / 2), self.ARROWSIZE),
                                   (tx + int(tw / 2) - x, 0),
                                   (tx + int(tw / 2) - x + int(self.ARROWSIZE / 2), self.ARROWSIZE),
                                   (w, self.ARROWSIZE),
                                   (w, h),
                                   (0, h)])
            self._sizer.SetDirection(direction)
        finally:
            memDC.SelectObject(wx.NullBitmap)
        self.SetDimensions(x, y, w, h)
        self.SetShape(wx.RegionFromBitmapColour(mask, wx.Colour(0, 0, 0)))
        self.Layout()


class BalloonTipManager(object):
    """
    Use this as a mixin in the top-level window that hosts balloon tip targets, to
    avoid them appearing all at once.
    """

    def __init__(self, *args, **kwargs):
        self.__tips = list()
        self.__displaying = None
        self.__kwargs = dict()
        self.__shutdown = False
        super(BalloonTipManager, self).__init__(*args, **kwargs)

        self.Bind(wx.EVT_CLOSE, self.__OnClose)

    def AddBalloonTip(self, target, message=None, title=None, bitmap=None, getRect=None, **kwargs):
        """Schedules a tip. Extra keyword arguments will be passed to L{OnBalloonTipShow} and L{OnBalloonTipClosed}."""
        for eTarget, eMessage, eTitle, eBitmap, eGetRect, eArgs in self.__tips:
            if (eTitle, eMessage) == (title, message):
                return
        self.__tips.append((target, message, title, bitmap, getRect, kwargs))
        self.__Try()

    def __Try(self):
        if self.__tips and not self.__shutdown and self.__displaying is None:
            target, message, title, bitmap, getRect, kwargs = self.__tips.pop(0)
            tip = BalloonTip(self, target, message=message, title=title, bitmap=bitmap, getRect=getRect)
            self.__displaying = tip
            self.OnBalloonTipShow(**kwargs)
            self.__kwargs = kwargs
            tip.Bind(wx.EVT_CLOSE, self.__OnCloseTip)

    def __OnClose(self, event):
        self.__shutdown = True
        event.Skip()

    def __OnCloseTip(self, event):
        event.Skip()
        self.__displaying = None
        self.OnBalloonTipClosed(**self.__kwargs)
        self.__Try()

    def OnBalloonTipShow(self, **kwargs):
        pass

    def OnBalloonTipClosed(self, **kwargs):
        pass


if __name__ == '__main__':
    class Frame(wx.Frame):
        def __init__(self):
            super(Frame, self).__init__(None, wx.ID_ANY, 'Test')

            self.btn = wx.Button(self, wx.ID_ANY, 'Show balloon')
            wx.EVT_BUTTON(self.btn, wx.ID_ANY, self.OnClick)
            s = wx.BoxSizer()
            s.Add(self.btn, 1, wx.EXPAND)
            self.SetSizer(s)
            self.Fit()

        def OnClick(self, event):
            BalloonTip(self, self.btn, '''Your bones don't break, mine do. That's clear. Your cells react to bacteria and viruses differently than mine. You don't get sick, I do. That's also clear. But for some reason, you and I react the exact same way to water. We swallow it too fast, we choke. We get some in our lungs, we drown. However unreal it may seem, we are connected, you and I. We're on the same curve, just on opposite ends.''', title='Title', bitmap=wx.ArtProvider.GetBitmap(wx.ART_TIP, wx.ART_MENU, (16, 16)))

    class App(wx.App):
        def OnInit(self):
            Frame().Show()
            return True

    App(0).MainLoop()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import config
from taskcoachlib.domain import task, date, effort
import test

 
class CommonTaskRelationshipManagerTestsMixin(object):
    def setUp(self):
        task.Task.settings = settings = config.Settings(load=False)
        now = self.now = date.Now()
        self.yesterday = now - date.ONE_DAY
        self.tomorrow = now + date.ONE_DAY
        self.parent = task.Task('parent')
        self.child = task.Task('child')
        self.parent.addChild(self.child)
        self.child.setParent(self.parent)
        self.child2 = task.Task('child2', plannedStartDateTime=now)
        self.grandchild = task.Task('grandchild', plannedStartDateTime=now)
        settings.set('behavior', 'markparentcompletedwhenallchildrencompleted', 
            str(self.markParentCompletedWhenAllChildrenCompleted))
        self.taskList = task.TaskList([self.parent, self.child2, 
                                       self.grandchild])
        
    # completion date
    
    def testMarkingOneOfTwoChildsCompletedNeverResultsInACompletedParent(self):
        self.parent.addChild(self.child2)
        self.child.setCompletionDateTime()
        self.failIf(self.parent.completed())

    def testMarkParentWithOneChildCompleted(self):
        self.parent.setCompletionDateTime()
        self.failUnless(self.child.completed())

    def testMarkParentWithTwoChildrenCompleted(self):
        self.parent.addChild(self.child2)        
        self.parent.setCompletionDateTime()
        self.failUnless(self.child.completed())
        self.failUnless(self.child2.completed())

    def testMarkParentNotCompleted(self):
        self.parent.setCompletionDateTime()
        self.failUnless(self.child.completed())
        self.parent.setCompletionDateTime(date.DateTime())
        self.failUnless(self.child.completed())

    def testMarkParentCompletedDoesNotChangeChildCompletionDate(self):
        self.parent.addChild(self.child2)        
        self.child.setCompletionDateTime(self.yesterday)
        self.parent.setCompletionDateTime()
        self.assertEqual(self.yesterday, self.child.completionDateTime())

    def testMarkChildNotCompleted(self):
        self.child.setCompletionDateTime()
        self.child.setCompletionDateTime(date.DateTime())
        self.failIf(self.parent.completed())
 
    def testAddCompletedChild(self):
        self.child2.setCompletionDateTime()
        self.parent.addChild(self.child2)
        self.failIf(self.parent.completed())

    def testAddUncompletedChild(self):
        self.child.setCompletionDateTime()
        self.parent.addChild(self.child2)
        self.failIf(self.parent.completed())
    
    def testAddUncompletedGrandchild(self):
        self.parent.setCompletionDateTime()
        self.child.addChild(self.grandchild)
        self.failIf(self.parent.completed())

    def testMarkParentCompletedYesterday(self):
        self.parent.setCompletionDateTime(self.yesterday)
        self.assertEqual(self.yesterday, self.child.completionDateTime())

    def testMarkTaskCompletedStopsEffortTracking(self):
        self.child.addEffort(effort.Effort(self.child))
        self.child.setCompletionDateTime()
        self.failIf(self.child.isBeingTracked())
    
    # recurrence
        
    def testMarkParentCompletedStopsChildRecurrence(self):
        self.child.setRecurrence(date.Recurrence('daily'))
        self.parent.setCompletionDateTime()
        self.failIf(self.child.recurrence())
        
    def testRecurringChildIsCompletedWhenParentIsCompleted(self):
        self.child.setRecurrence(date.Recurrence('daily'))
        self.parent.setCompletionDateTime()
        self.failUnless(self.child.completed())
        
    def shouldMarkCompletedWhenAllChildrenCompleted(self, parent):
        return parent.shouldMarkCompletedWhenAllChildrenCompleted() == True or \
            (parent.shouldMarkCompletedWhenAllChildrenCompleted() == None and \
             self.markParentCompletedWhenAllChildrenCompleted == True)
        
    def testMarkLastChildCompletedMakesParentRecur(self):
        self.parent.setPlannedStartDateTime(self.now)
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.child.setCompletionDateTime(self.now)
        expectedPlannedStartDateTime = self.now
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            expectedPlannedStartDateTime += date.TimeDelta(days=7)
        self.assertAlmostEqual(expectedPlannedStartDateTime.toordinal(), 
                               self.parent.plannedStartDateTime().toordinal())

    def testMarkLastChildCompletedMakesParentRecur_AndThusChildToo(self):
        self.child.setPlannedStartDateTime(self.now)
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.parent.setPlannedStartDateTime(self.now)
        self.child.setCompletionDateTime(self.now)
        expectedPlannedStartDateTime = self.now        
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            expectedPlannedStartDateTime += date.TimeDelta(days=7)
        self.assertAlmostEqual(expectedPlannedStartDateTime.toordinal(), 
                               self.child.plannedStartDateTime().toordinal())

    def testMarkLastChildCompletedMakesParentRecur_AndThusChildIsNotCompleted(self):
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.child.setCompletionDateTime()
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            self.failIf(self.child.completed())
        else:
            self.failUnless(self.child.completed())

    def testMarkLastGrandChildCompletedMakesParentRecur(self):
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.parent.setPlannedStartDateTime(self.now)
        self.child.addChild(self.grandchild)
        self.grandchild.setParent(self.child)
        self.grandchild.setCompletionDateTime(self.now)
        expectedPlannedStartDateTime = self.now
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            expectedPlannedStartDateTime += date.TimeDelta(days=7)
        self.assertAlmostEqual(expectedPlannedStartDateTime.toordinal(), 
                               self.parent.plannedStartDateTime().toordinal())

    def testMarkLastGrandChildCompletedMakesParentRecur_AndThusGrandChildToo(self):
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.child.addChild(self.grandchild)
        self.grandchild.setParent(self.child)
        self.grandchild.setCompletionDateTime(self.now)
        expectedPlannedStartDateTime = self.now
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            expectedPlannedStartDateTime += date.TimeDelta(days=7)
        self.assertAlmostEqual(expectedPlannedStartDateTime.toordinal(), 
                               self.grandchild.plannedStartDateTime().toordinal())

    def testMarkLastChildCompletedMakesParentRecur_AndThusGrandChildIsNotCompleted(self):
        self.parent.setRecurrence(date.Recurrence('weekly'))
        self.child.addChild(self.grandchild)
        self.grandchild.setParent(self.child)
        self.grandchild.setCompletionDateTime()
        if self.shouldMarkCompletedWhenAllChildrenCompleted(self.parent):
            self.failIf(self.grandchild.completed())
        else:
            self.failUnless(self.grandchild.completed())


class MarkParentTaskCompletedTestsMixin(object):
    ''' Tests where we expect to parent task to be marked completed, based on
        the fact that all children are completed. This happens when the global
        setting is on and task is indifferent or the task specific setting is 
        on. '''
        
    def testMarkOnlyChildCompleted(self):
        self.child.setCompletionDateTime()
        self.failUnless(self.parent.completed())
        
    def testMarkOnlyGrandchildCompleted(self):
        self.child.addChild(self.grandchild)
        self.grandchild.setCompletionDateTime()
        self.failUnless(self.parent.completed())                        
              
    def testAddCompletedChildAsOnlyChild(self):
        self.grandchild.setCompletionDateTime()
        self.child.addChild(self.grandchild)
        self.failUnless(self.child.completed())
        
    def testMarkChildCompletedYesterday(self):    
        self.child.setCompletionDateTime(self.yesterday)
        self.assertEqual(self.yesterday, self.parent.completionDateTime())
        
    def testRemoveLastUncompletedChild(self):
        self.parent.addChild(self.child2)
        self.child.setCompletionDateTime()
        self.parent.removeChild(self.child2)
        self.failUnless(self.parent.completed())
    
    
class DontMarkParentTaskCompletedTestsMixin(object):
    ''' Tests where we expect the parent task not to be marked completed when 
        all children are completed. This should be the case when the global
        setting is off and task is indifferent or when the task specific 
        setting is off. '''
  
    def testMarkOnlyChildCompletedDoesNotMarkParentCompleted(self):
        self.child.setCompletionDateTime()
        self.failIf(self.parent.completed())

    def testMarkOnlyGrandchildCompletedDoesNotMarkParentCompleted(self):
        self.child.addChild(self.grandchild)
        self.grandchild.setCompletionDateTime()
        self.failIf(self.parent.completed())    
 
    def testAddCompletedChildAsOnlyChildDoesNotMarkParentCompleted(self):
        self.grandchild.setCompletionDateTime()
        self.child.addChild(self.grandchild)
        self.failIf(self.child.completed())

    def testMarkChildCompletedYesterdayDoesNotAffectParentCompletionDate(self):    
        self.child.setCompletionDateTime(self.yesterday)
        self.assertEqual(date.DateTime(), self.parent.completionDateTime())

    def testRemoveLastUncompletedChildDoesNotMarkParentCompleted(self):
        self.parent.addChild(self.child2)
        self.child.setCompletionDateTime()
        self.parent.removeChild(self.child2)
        self.failIf(self.parent.completed())        


class MarkParentCompletedAutomaticallyIsOn(CommonTaskRelationshipManagerTestsMixin,
                                           MarkParentTaskCompletedTestsMixin,
                                           test.TestCase):
    markParentCompletedWhenAllChildrenCompleted = True


class MarkParentCompletedAutomaticallyIsOff(CommonTaskRelationshipManagerTestsMixin,
                                            DontMarkParentTaskCompletedTestsMixin,
                                            test.TestCase):
    markParentCompletedWhenAllChildrenCompleted = False
              

class MarkParentCompletedAutomaticallyIsOnButTaskSettingIsOff( \
        CommonTaskRelationshipManagerTestsMixin, test.TestCase,
        DontMarkParentTaskCompletedTestsMixin):
    markParentCompletedWhenAllChildrenCompleted = True
    
    def setUp(self):
        super(MarkParentCompletedAutomaticallyIsOnButTaskSettingIsOff, self).setUp()
        for eachTask in self.parent, self.child:
            eachTask.setShouldMarkCompletedWhenAllChildrenCompleted(False)


class MarkParentCompletedAutomaticallyIsOffButTaskSettingIsOn( \
        CommonTaskRelationshipManagerTestsMixin, test.TestCase,
        MarkParentTaskCompletedTestsMixin):
    markParentCompletedWhenAllChildrenCompleted = False

    def setUp(self):
        super(MarkParentCompletedAutomaticallyIsOffButTaskSettingIsOn, self).setUp()
        for eachTask in self.parent, self.child:
            eachTask.setShouldMarkCompletedWhenAllChildrenCompleted(True)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain import base
from taskcoachlib import patterns


class CategorizableContainer(base.Collection):
    @patterns.eventSource
    def extend(self, items, event=None):
        super(CategorizableContainer, self).extend(items, event=event)
        for item in self._compositesAndAllChildren(items):
            for category in item.categories():
                category.addCategorizable(item, event=event)

    @patterns.eventSource                
    def removeItems(self, items, event=None):
        super(CategorizableContainer, self).removeItems(items, event=event)
        for item in self._compositesAndAllChildren(items):
            for category in item.categories():
                category.removeCategorizable(item, event=event)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.domain import date, task
from taskcoachlib.thirdparty.pubsub import pub
import composite
import effortlist
import effort


class EffortAggregator(patterns.SetDecorator, 
                       effortlist.EffortUICommandNamesMixin):
    ''' This class observes an TaskList and aggregates the individual effort
        records to CompositeEfforts, e.g. per day or per week. Whenever a 
        CompositeEffort becomes empty, for example because effort is deleted,
        it sends an 'empty' event so that the aggregator can remove the
        (now empty) CompositeEffort from itself. '''
        
    def __init__(self, *args, **kwargs):
        self.__composites = {}
        self.__trackedComposites = set()
        aggregation = kwargs.pop('aggregation')
        assert aggregation in ('day', 'week', 'month')
        aggregation = aggregation.capitalize()
        self.__start_of_period = getattr(date.DateTime, 'startOf%s' % aggregation)
        self.__end_of_period = getattr(date.DateTime, 'endOf%s' % aggregation)
        super(EffortAggregator, self).__init__(*args, **kwargs)
        pub.subscribe(self.onCompositeEmpty, 
                      composite.CompositeEffort.compositeEmptyEventType())
        pub.subscribe(self.onTaskEffortChanged, 
                      task.Task.effortsChangedEventType())
        patterns.Publisher().registerObserver(self.onChildAddedToTask,
            eventType=task.Task.addChildEventType())
        patterns.Publisher().registerObserver(self.onChildRemovedFromTask,
            eventType=task.Task.removeChildEventType())
        patterns.Publisher().registerObserver(self.onTaskRemoved, 
                                              self.observable().removeItemEventType(),
                                              eventSource=self.observable())
        pub.subscribe(self.onEffortStartChanged, 
                      effort.Effort.startChangedEventType())
        pub.subscribe(self.onRevenueChanged,
                      task.Task.hourlyFeeChangedEventType())

    def detach(self):
        super(EffortAggregator, self).detach()
        patterns.Publisher().removeObserver(self.onChildAddedToTask)
        patterns.Publisher().removeObserver(self.onChildRemovedFromTask)
        patterns.Publisher().removeObserver(self.onTaskRemoved)

    def extend(self, efforts):  # pylint: disable=W0221
        for effort in efforts:
            effort.task().addEffort(effort)

    def removeItems(self, efforts):  # pylint: disable=W0221
        for effort in efforts:
            effort.task().removeEffort(effort)
            
    @patterns.eventSource
    def extendSelf(self, tasks, event=None):
        ''' extendSelf is called when an item is added to the observed
            list. The default behavior of extendSelf is to add the item
            to the observing list (i.e. this list) unchanged. We override 
            the default behavior to first get the efforts from the task
            and then group the efforts by time period. '''
        new_composites = []
        for task in tasks:  # pylint: disable=W0621
            new_composites.extend(self.__create_composites(task, task.efforts()))
        self.__extend_self_with_composites(new_composites, event=event)
        
    @patterns.eventSource
    def __extend_self_with_composites(self, new_composites, event=None):
        ''' Add composites to the aggregator. '''
        super(EffortAggregator, self).extendSelf(new_composites, event=event)
        for new_composite in new_composites:
            if new_composite.isBeingTracked():
                self.__trackedComposites.add(new_composite)
                pub.sendMessage(effort.Effort.trackingChangedEventType(),
                                newValue=True, sender=new_composite)

    @patterns.eventSource
    def removeItemsFromSelf(self, tasks, event=None):
        ''' removeItemsFromSelf is called when an item is removed from the 
            observed list. The default behavior of removeItemsFromSelf is to 
            remove the item from the observing list (i.e. this list)
            unchanged. We override the default behavior to remove the 
            tasks' efforts from the CompositeEfforts they are part of. '''
        composites_to_remove = []
        for task in tasks:  # pylint: disable=W0621
            composites_to_remove.extend(self.__composites_to_remove(task))
        self.__remove_composites_from_self(composites_to_remove, event=event)
         
    @patterns.eventSource
    def __remove_composites_from_self(self, composites_to_remove, event=None):
        ''' Remove composites from the aggregator. '''
        self.__trackedComposites.difference_update(set(composites_to_remove))
        super(EffortAggregator, self).removeItemsFromSelf(composites_to_remove, 
                                                          event=event)
        
    def onTaskRemoved(self, event):
        ''' Whenever tasks are removed, find the composites that 
            (did) contain effort of those tasks and update them. '''
        affected_composites = self.__get_composites_for_tasks(event.values())
        for affected_composite in affected_composites:
            affected_composite._invalidateCache()
            affected_composite.notifyObserversOfDurationOrEmpty()
            
    def onTaskEffortChanged(self, newValue, sender):
        if sender not in self.observable():
            return
        new_composites = []
        newValue, oldValue = newValue
        efforts_added = [effort for effort in newValue if effort not in oldValue]
        efforts_removed = [effort for effort in oldValue if effort not in newValue]
        new_composites.extend(self.__create_composites(sender, efforts_added))
        self.__extend_self_with_composites(new_composites)
        for affected_composite in self.__get_composites_for_efforts(efforts_added + efforts_removed):
            is_tracked = affected_composite.isBeingTracked()
            was_tracked = affected_composite in self.__trackedComposites
            if is_tracked and not was_tracked:
                self.__trackedComposites.add(affected_composite)
            elif not is_tracked and was_tracked:
                self.__trackedComposites.remove(affected_composite)
            affected_composite.onTimeSpentChanged(newValue, sender)
        
    def onChildAddedToTask(self, event):
        new_composites = []
        for task in event.sources():  # pylint: disable=W0621
            if task in self.observable():
                child = event.value(task)
                new_composites.extend(self.__create_composites(task,
                    child.efforts(recursive=True)))
        self.__extend_self_with_composites(new_composites)
        
    def onChildRemovedFromTask(self, event):
        affected_composites = self.__get_composites_for_tasks(event.sources() | set(event.values()))
        for affected_composite in affected_composites:
            affected_composite._invalidateCache()
            affected_composite.notifyObserversOfDurationOrEmpty()

    def onCompositeEmpty(self, sender):
        # pylint: disable=W0621
        if sender not in self:
            return
        key = self.__key_for_composite(sender)
        if key in self.__composites:
            # A composite may already have been removed, e.g. when a
            # parent and child task have effort in the same period
            del self.__composites[key]
        self.__remove_composites_from_self([sender])
        
    def onEffortStartChanged(self, newValue, sender):  # pylint: disable=W0613
        new_composites = []
        key = self.__key_for_effort(sender)
        task = sender.task()  # pylint: disable=W0621
        if (task in self.observable()) and (key not in self.__composites):
            new_composites.extend(self.__create_composites(task, [sender]))
        self.__extend_self_with_composites(new_composites)
        for affected_composite in self.__get_composites_for_efforts([sender]):
            is_tracked = affected_composite.isBeingTracked()
            was_tracked = affected_composite in self.__trackedComposites
            if is_tracked and not was_tracked:
                self.__trackedComposites.add(affected_composite)
            elif not is_tracked and was_tracked:
                self.__trackedComposites.remove(affected_composite)
            affected_composite.onTimeSpentChanged(newValue, sender)
            
    def onRevenueChanged(self, newValue, sender):
        for affected_composite in self.__get_composites_for_tasks([sender]):
            affected_composite.onRevenueChanged(newValue, sender)
            
    def __get_composites_for_tasks(self, tasks):
        tasks = set(tasks)
        return [each_composite for each_composite in self \
                if each_composite.task() in tasks or \
                (each_composite.task().__class__.__name__ == 'Total' and \
                 tasks & each_composite.tasks())]
        
    def __get_composites_for_efforts(self, efforts):
        efforts = set(efforts)
        return [each_composite for each_composite in self \
                if set(each_composite._getEfforts()) & efforts]
        
    def __create_composites(self, task, efforts):  # pylint: disable=W0621
        new_composites = []
        for effort in efforts:
            new_composites.extend(self.__create_composites_for_task(effort, task))
            new_composites.extend(self.__create_composite_for_period(effort))
        return new_composites

    def __create_composites_for_task(self, an_effort, task):  # pylint: disable=W0621
        new_composites = []
        for each_task in [task] + task.ancestors():
            key = self.__key_for_effort(an_effort, each_task)
            if key in self.__composites:
                self.__composites[key].addEffort(an_effort)
                continue
            new_composite = composite.CompositeEffort(*key)  # pylint: disable=W0142
            new_composite.addEffort(an_effort)
            self.__composites[key] = new_composite
            new_composites.append(new_composite)
        return new_composites
    
    def __create_composite_for_period(self, an_effort):
        key = self.__key_for_period(an_effort)
        if key in self.__composites:
            self.__composites[key].addEffort(an_effort)
            return []
        new_composite_per_period = composite.CompositeEffortPerPeriod(key[0], 
                                          key[1], self.observable(), an_effort)
        self.__composites[key] = new_composite_per_period
        return [new_composite_per_period]

    def __composites_to_remove(self, task):  # pylint: disable=W0621
        efforts = task.efforts()
        task_and_ancestors = [task] + task.ancestors()
        composites_to_remove = []
        for effort in efforts:
            for task in task_and_ancestors:
                composites_to_remove.extend(self.__composite_to_remove(effort, task))
        return composites_to_remove
        
    def __composite_to_remove(self, an_effort, task):  # pylint: disable=W0613,W0621
        key = self.__key_for_effort(an_effort, task)
        # A composite may already have been removed, e.g. when a
        # parent and child task have effort in the same period
        return [self.__composites.pop(key)] if key in self.__composites else []

    def maxDateTime(self):
        stop_times = [effort.getStop() for composite_effort in self for effort
                      in composite_effort if effort.getStop() is not None]
        return max(stop_times) if stop_times else None

    @staticmethod
    def __key_for_composite(composite_effort):
        if composite_effort.task().__class__.__name__ == 'Total':
            return (composite_effort.getStart(), composite_effort.getStop())
        else:
            return (composite_effort.task(), composite_effort.getStart(), 
                    composite_effort.getStop())
    
    def __key_for_effort(self, effort, task=None):  # pylint: disable=W0621
        task = task or effort.task()
        effort_start = effort.getStart()
        return (task, self.__start_of_period(effort_start), 
                      self.__end_of_period(effort_start))
        
    def __key_for_period(self, effort):
        key = self.__key_for_effort(effort)
        return key[1], key[2]
    
    @classmethod
    def sortEventType(class_):
        return 'this event type is not used'  # pragma: no cover



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, wx
from taskcoachlib import gui, config, persistence
from taskcoachlib.domain import category
 

class CategoryViewerTest(test.wxTestCase):
    def setUp(self):
        super(CategoryViewerTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.categories = self.taskFile.categories()
        self.viewer = gui.viewer.CategoryViewer(self.frame, self.taskFile, 
                                                self.settings)

    def tearDown(self):
        super(CategoryViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def addTwoCategories(self):
        cat1 = category.Category('1')
        cat2 = category.Category('2')
        self.categories.extend([cat2, cat1])
        return cat1, cat2
        
    def testInitialSize(self):
        self.assertEqual(0, self.viewer.size())

    def testCopyCategoryWithChildren(self):
        parent, child = self.addTwoCategories()
        parent.addChild(child)
        copy = parent.copy()
        self.categories.append(copy)
        self.viewer.expandAll()
        self.assertEqual(4, self.viewer.size())

    def testSortInWidget(self):
        self.addTwoCategories()
        widget = self.viewer.widget
        for item, cat in zip(widget.GetItemChildren(), self.viewer.presentation()):
            self.assertEqual(cat.subject(), widget.GetItemText(item))
            
    def testSelectAll(self):
        self.addTwoCategories()
        self.viewer.widget.select_all()
        self.viewer.updateSelection()
        self.assertEqual(2, len(self.viewer.curselection()))
        
    def testFilterOnAllCheckedCategoriesSetsSetting(self):
        self.viewer.filterUICommand.doChoice(True)
        self.failUnless(self.settings.getboolean('view', 'categoryfiltermatchall'))

    def testFilterOnAnyCheckedCategoriesSetsSetting(self):
        self.viewer.filterUICommand.doChoice(False)
        self.failIf(self.settings.getboolean('view', 'categoryfiltermatchall'))
    


# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import gui, config, persistence, operating_system
from taskcoachlib.domain import task, effort, date, note, attachment
from taskcoachlib.gui import uicommand
from unittests import dummy
import test
import wx


class TaskEditorSetterMixin(object):
    def setSubject(self, newSubject):
        page = self.editor._interior[0]
        page._subjectEntry.SetFocus()
        page._subjectEntry.SetValue(newSubject)
        return page

    def setDescription(self, newDescription):
        page = self.editor._interior[0]
        page._descriptionEntry.SetFocus()
        page._descriptionEntry.SetValue(newDescription)
        return page
    
    def setPlannedStartDateTime(self, dateTime):
        self.setDateTime(self.editor._interior[1]._plannedStartDateTimeEntry, 
                         dateTime)
        
    def setDueDateTime(self, dateTime):
        self.setDateTime(self.editor._interior[1]._dueDateTimeEntry, dateTime)

    def setActualStartDateTime(self, dateTime):
        self.setDateTime(self.editor._interior[1]._actualStartDateTimeEntry, 
                         dateTime)

    def setCompletionDateTime(self, dateTime):
        self.setDateTime(self.editor._interior[1]._completionDateTimeEntry, 
                         dateTime)

    def setReminder(self, dateTime):
        self.setDateTime(self.editor._interior[1]._reminderDateTimeEntry, 
                         dateTime)
        
    def setDateTime(self, entry, dateTime):
        entry.SetValue(dateTime)
        entry.onDateTimeCtrlEdited()
        wx.YieldIfNeeded()
        
    def setRecurrence(self, newRecurrence):
        recurrenceEntry = self.editor._interior[1]._recurrenceEntry
        recurrenceEntry.SetValue(newRecurrence)
        recurrenceEntry.onRecurrenceEdited()
        wx.YieldIfNeeded()


class TaskEditorBySettingFocusMixin(TaskEditorSetterMixin):
    def setSubject(self, newSubject):
        page = super(TaskEditorBySettingFocusMixin, self).setSubject(newSubject)
        if operating_system.isGTK(): 
            page._subjectSync.onAttributeEdited(dummy.Event())  # pragma: no cover
        else:
            page._descriptionEntry.SetFocus()  # pragma: no cover
        
    def setDescription(self, newDescription):
        page = super(TaskEditorBySettingFocusMixin, self).setDescription(newDescription)
        if operating_system.isGTK(): 
            page._descriptionSync.onAttributeEdited(dummy.Event())  # pragma: no cover
        else:
            page._subjectEntry.SetFocus()  # pragma: no cover


class TaskEditBookWithPerspective(gui.dialog.editor.TaskEditBook):
    testPerspective = '503e3c7d0000018300000002=+0,1,2,3,4,5,6,7,8,9@layout2|name=dummy;caption=;state=67372030;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=381;besth=173;minw=381;minh=173;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=503e3c7d0000018300000002;caption=;state=67372028;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=200;besth=200;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=202|'

    def perspective(self):
        return self.testPerspective


class TaskEditorWithPerspective(gui.dialog.editor.TaskEditor):
    EditBookClass = TaskEditBookWithPerspective


class TaskEditorTestCase(test.wxTestCase):
    extraSettings = list()
    editorClass = gui.dialog.editor.TaskEditor

    def setUp(self):
        super(TaskEditorTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        for section, name, value in self.extraSettings:
            self.settings.set(section, name, value)
        self.today = date.Now()
        self.tomorrow = self.today + date.ONE_DAY
        self.yesterday = self.today - date.ONE_DAY
        self.twodaysago = self.yesterday - date.ONE_DAY
        self.taskFile = persistence.TaskFile()
        self.taskList = self.taskFile.tasks()
        self.taskList.extend(self.createTasks())
        self.editor = self.editorClass(self.frame, self.getItems(),
            self.settings, self.taskList, self.taskFile)

    def tearDown(self):
        # TaskEditor uses CallAfter for setting the focus, make sure those 
        # calls are dealt with, otherwise they'll turn up in other tests
        if operating_system.isGTK():
            wx.Yield()  # pragma: no cover 
        super(TaskEditorTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()
        
    def createTasks(self):
        raise NotImplementedError  # pragma: no cover
    
    def getItems(self):
        raise NotImplementedError  # pragma: no cover


class EditorDisplayTest(TaskEditorTestCase):
    ''' Does the editor display the task data correctly when opened? '''
    
    def getItems(self):
        return [self.task]
    
    def createTasks(self):
        # pylint: disable=W0201
        self.task = task.Task('Task to edit')
        self.stop_datetime = date.DateTime(2012, 12, 12, 12, 12)
        self.task.setRecurrence( \
            date.Recurrence('daily', amount=1, 
                            stop_datetime=self.stop_datetime))
        return [self.task]
    
    def testSubject(self):
        self.assertEqual('Task to edit',
                         self.editor._interior[0]._subjectEntry.GetValue())

    def testDueDateTime(self):
        self.assertEqual(date.DateTime(),
                         self.editor._interior[1]._dueDateTimeEntry.GetValue())
 
    def testActualStartDateTime(self):
        self.assertEqual(date.DateTime(),
                         self.editor._interior[1]._actualStartDateTimeEntry.GetValue())
       
    def testRecurrenceUnit(self):
        choice = self.editor._interior[1]._recurrenceEntry._recurrencePeriodEntry
        self.assertEqual('Daily', choice.GetString(choice.GetSelection()))

    def testRecurrenceFrequency(self):
        freq = self.editor._interior[1]._recurrenceEntry._recurrenceFrequencyEntry
        self.assertEqual(1, freq.GetValue())    

    def testRecurrenceStopDateTime(self):
        stop = self.editor._interior[1]._recurrenceEntry._recurrenceStopDateTimeEntry
        self.assertEqual(self.stop_datetime, stop.GetValue())    


class EditTaskTestMixin(object):
    def getItems(self):
        return [self.task]

    def createTasks(self):
        # pylint: disable=W0201
        self.task = task.Task('Task to edit')
        self.attachment = attachment.FileAttachment('some attachment')
        self.task.addAttachments(self.attachment)  # pylint: disable=E1101
        return [self.task]

    def testEditSubject(self):
        self.setSubject('Done')
        self.assertEqual('Done', self.task.subject())

    def testEditDescription(self):
        self.setDescription('Description')
        self.assertEqual('Description', self.task.description())

    # pylint: disable=W0212
   
    def testSetPlannedStartDateTime(self):
        self.setPlannedStartDateTime(self.tomorrow)
        self.assertAlmostEqual(self.tomorrow.toordinal(), 
                               self.task.plannedStartDateTime().toordinal(),
                               places=2)

    def testSetDueDateTime(self):
        self.setDueDateTime(self.tomorrow)
        self.assertAlmostEqual(self.tomorrow.toordinal(), 
                               self.task.dueDateTime().toordinal(),
                               places=2)

    def testSetActualStartDateTime(self):
        self.setActualStartDateTime(self.tomorrow)
        self.assertAlmostEqual(self.tomorrow.toordinal(), 
                               self.task.actualStartDateTime().toordinal(),
                               places=2)

    def testSetCompletionDateTime(self):
        self.setCompletionDateTime(self.tomorrow)
        self.assertAlmostEqual(self.tomorrow.toordinal(), 
                               self.task.completionDateTime().toordinal(),
                               places=2)

    def testSetUncompleted(self):
        self.setCompletionDateTime(date.Now())
        self.setCompletionDateTime(date.DateTime())
        self.assertEqual(date.DateTime(), self.task.completionDateTime())

    def testSetReminder(self):
        reminderDateTime = date.DateTime(2005,1,1)
        self.setReminder(reminderDateTime)
        self.assertEqual(reminderDateTime, self.task.reminder())

    def testSetRecurrence(self):
        self.setRecurrence(date.Recurrence('weekly'))
        self.assertEqual('weekly', self.task.recurrence().unit)
        
    def testSetDailyRecurrence(self):
        self.setRecurrence(date.Recurrence('daily', amount=1))
        self.assertEqual('daily', self.task.recurrence().unit)
        self.assertEqual(1, self.task.recurrence().amount)
        
    def testSetYearlyRecurrence(self):
        self.setRecurrence(date.Recurrence('yearly'))
        self.assertEqual('yearly', self.task.recurrence().unit)
        
    def testSetMaxRecurrence(self):
        self.setRecurrence(date.Recurrence('weekly', maximum=10))
        self.assertEqual(10, self.task.recurrence().max)

    def testSetRecurrenceStopDateTime(self):
        stop = date.DateTime(2012, 3, 4, 10, 0)
        self.setRecurrence(date.Recurrence('weekly', stop_datetime=stop))
        self.assertEqual(stop, self.task.recurrence().stop_datetime)
        
    def testSetRecurrenceFrequency(self):
        self.setRecurrence(date.Recurrence('weekly', amount=3))
        self.assertEqual(3, self.task.recurrence().amount)
        
    def testSetRecurrenceSameWeekday(self):
        self.setRecurrence(date.Recurrence('monthly', sameWeekday=True))
        self.failUnless(self.task.recurrence().sameWeekday)
        
    def testPriority(self):
        self.editor._interior[0]._priorityEntry.SetValue(45)
        self.assertEqual(45, self.editor._interior[0]._priorityEntry.GetValue())
        
    def testSetNegativePriority(self):
        self.editor._interior[0]._priorityEntry.SetValue(-1)
        self.editor._interior[0]._prioritySync.onAttributeEdited(dummy.Event())
        self.assertEqual(-1, self.task.priority())
        
    def testSetHourlyFee(self):
        self.editor._interior[5]._hourlyFeeEntry.SetValue(100)
        self.editor._interior[5]._hourlyFeeSync.onAttributeEdited(dummy.Event())
        self.assertEqual(100, self.task.hourlyFee())

    def testSetFixedFee(self):
        self.editor._interior[5]._fixedFeeEntry.SetValue(100.5)
        self.editor._interior[5]._fixedFeeSync.onAttributeEdited(dummy.Event())
        self.assertEqual(100.5, self.task.fixedFee())

    def testBehaviorMarkCompleted(self):
        page = self.editor._interior[3]
        page._shouldMarkCompletedEntry.SetStringSelection('Yes')
        page._shouldMarkCompletedSync.onAttributeEdited(dummy.Event())
        self.assertEqual(True, 
                         self.task.shouldMarkCompletedWhenAllChildrenCompleted())

    def testAddAttachment(self):
        self.editor._interior[8].viewer.onDropFiles(self.task, ['filename'])
        # pylint: disable=E1101
        self.failUnless('filename' in [att.location() for att in self.task.attachments()])
        self.failUnless('filename' in [att.subject() for att in self.task.attachments()])
        
    def testRemoveAttachment(self):
        self.editor._interior[8].viewer.select(self.task.attachments())
        self.editor._interior[8].viewer.deleteItemCommand().do()
        self.assertEqual([], self.task.attachments())  # pylint: disable=E1101

    def testOpenAttachmentWithNonAsciiFileName(self):
        self.errorMessage = ''  # pylint: disable=W0201
        
        def onError(*args, **kwargs):  # pylint: disable=W0613
            self.errorMessage = args[0]  # pragma: no cover
            
        att = attachment.FileAttachment(u'tÃƒÂ©st.ÃƒÂ©')
        openAttachment = uicommand.AttachmentOpen(\
            viewer=self.editor._interior[6].viewer,
            attachments=attachment.AttachmentList([att]),
            settings=self.settings)
        openAttachment.doCommand(None, showerror=onError)
        self.failIf(self.errorMessage)

    def testAddNote(self):
        viewer = self.editor._interior[7].viewer
        viewer.newItemCommand(viewer.presentation()).do()
        self.assertEqual(1, len(self.task.notes()))
        
    def testAddNoteWithSubnote(self):
        parent = note.Note(subject='New note')
        child = note.Note(subject='Child')
        parent.addChild(child)
        child.setParent(parent)
        viewer = self.editor._interior[7].viewer
        viewer.newItemCommand(viewer.presentation()).do()
        viewer.newSubItemCommandClass()(list=viewer.presentation(), 
                                        items=viewer.presentation()).do()
        # Only the parent note should be added to the notes list:
        self.assertEqual(1, len(self.task.notes())) 

    def testNewNote(self):
        viewer = self.editor._interior[7].viewer
        wx.GetApp().TopWindow.taskFile = self.taskFile
        command = gui.uicommand.NoteNew(notes=viewer.presentation(),
                                        settings=self.settings,
                                        viewer=viewer)
        dialog = command.doCommand(None, show=False)
        dialog.ok()
        self.assertEqual(1, len(self.task.notes()))


class EditTaskTestBySettingFocus(TaskEditorBySettingFocusMixin, 
                                 EditTaskTestMixin, TaskEditorTestCase):
    pass


class EditTaskWithChildrenMixin(object):
    def getItems(self):
        return [self.parent]

    def createTasks(self):
        # pylint: disable=W0201
        self.parent = task.Task('Parent', plannedStartDateTime=date.Now())
        self.child = task.Task('Child', plannedStartDateTime=date.Now())
        self.parent.addChild(self.child)
        return [self.parent]  # self.child is added to tasklist automatically

    def testEditSubject(self):
        self.setSubject('New Parent Subject')
        self.assertEqual('New Parent Subject', self.parent.subject())


class EditTaskWithChildrenTestBySettingFocus(TaskEditorBySettingFocusMixin, 
                                             EditTaskWithChildrenMixin, 
                                             TaskEditorTestCase):
    pass


class EditTaskWithEffortTest(TaskEditorTestCase):    
    def getItems(self):
        return [self.task]

    def createTasks(self):
        self.task = task.Task('task')  # pylint: disable=W0201
        self.task.addEffort(effort.Effort(self.task))
        return [self.task]
    
    def testEffortIsShown(self):
        self.assertEqual(1, 
                         self.editor._interior[6].viewer.widget.GetItemCount())
                           
        
class FocusTest(TaskEditorTestCase):
    def createTasks(self):
        self.task = task.Task('Task to edit')
        return [self.task]
    
    def getItems(self):
        return [self.task]

    def testFocus(self):
        # pylint: disable=W0212
        self.assertEqual(self.editor._interior[0]._subjectEntry, 
                         wx.Window_FindFocus())

    def testSelection(self):
        # pylint: disable=W0212
        self.assertEqual(self.editor._interior[0]._subjectEntry.GetStringSelection(),
                         self.editor._interior[0]._subjectEntry.GetValue())


class FocusTestWithGTKSetting(TaskEditorTestCase):
    extraSettings = [('os_linux', 'focustextentry', 'False')]
    def createTasks(self):
        self.task = task.Task('Task to edit')
        return [self.task]
    
    def getItems(self):
        return [self.task]

    def testFocus(self):
        # pylint: disable=W0212
        self.assertNotEqual(self.editor._interior[0]._subjectEntry, 
                            wx.Window_FindFocus())

    def testSelection(self):
        # pylint: disable=W0212
        self.assertEqual(self.editor._interior[0]._subjectEntry.GetStringSelection(), u'')


class FocusTestWithPerspective(FocusTest):
    editorClass = TaskEditorWithPerspective


class FocusTestWithPerspectiveAndGTKSetting(FocusTestWithGTKSetting):
    editorClass = TaskEditorWithPerspective


class DatesTestBase(TaskEditorSetterMixin, TaskEditorTestCase):
    def createTasks(self):
        # pylint: disable=W0201
        self.task = task.Task('Task to edit')
        return [self.task]

    def getItems(self):
        return [self.task]


class DatesStartDueTest(DatesTestBase):
    extraSettings = [('view', 'datestied', 'startdue')]

    def testChangePlannedStartDateChangesDueDate(self):
        self.setPlannedStartDateTime(self.yesterday)
        self.setDueDateTime(self.today)
        self.setPlannedStartDateTime(self.today)
        self.assertAlmostEqual(self.editor._interior[1]._dueDateTimeEntry.GetValue().toordinal(),
                               self.tomorrow.toordinal(),
                               places=2)


class DatesDueStartBase(DatesTestBase):
    extraSettings = [('view', 'datestied', 'duestart')]

    def testChangeDueDateChangesPlannedStartDate(self):
        self.setPlannedStartDateTime(self.yesterday)
        self.setDueDateTime(self.today)
        self.setDueDateTime(self.yesterday)
        self.assertAlmostEqual(self.editor._interior[1]._plannedStartDateTimeEntry.GetValue().toordinal(),
                               self.twodaysago.toordinal(),
                               places=2)


class DatesTest(DatesTestBase):
    def testChangePlannedStartDateDoesNotChangeDueDate(self):
        self.setPlannedStartDateTime(self.yesterday)
        self.setDueDateTime(self.today)
        self.setPlannedStartDateTime(self.today)
        self.assertAlmostEqual(self.editor._interior[1]._dueDateTimeEntry.GetValue().toordinal(),
                               self.today.toordinal(),
                               places=2)

    def testChangeDueDateDoesNotChangePlannedStartDate(self):
        self.setPlannedStartDateTime(self.yesterday)
        self.setDueDateTime(self.today)
        self.setDueDateTime(self.yesterday)
        self.assertAlmostEqual(self.editor._interior[1]._plannedStartDateTimeEntry.GetValue().toordinal(),
                               self.yesterday.toordinal(),
                               places=2)
